import unittest

import requests_mock
from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import Request

from perimeterx import px_constants
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext
from perimeterx.px_proxy import PxProxy


class Test_PXProxy(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.config = PxConfig({'app_id': 'PXfake_app_id'})
        cls.headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'content_length': '100'}


    def test_should_reverse_request(self):
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/init.js')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        px_proxy = PxProxy(self.config)

        should_reverse = px_proxy.should_reverse_request(context.uri)
        self.assertTrue(should_reverse)
        should_reverse = px_proxy.should_reverse_request(context.uri)
        self.assertTrue(should_reverse)
        should_reverse = px_proxy.should_reverse_request(context.uri)
        self.assertTrue(should_reverse)

    @requests_mock.mock()
    def test_send_reverse_client_request(self, mock):
        content = 'client js content'
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/init.js')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        headers = {'host': px_constants.CLIENT_HOST,
                   px_constants.FIRST_PARTY_HEADER: '1',
                   px_constants.ENFORCER_TRUE_IP_HEADER: context.ip,
                   px_constants.FIRST_PARTY_FORWARDED_FOR: '127.0.0.1'}
        mock.get(url='https://client.perimeterx.net/PXfake_app_id/main.min.js', text=content, request_headers=headers,
                 status_code=200, reason='OK')
        px_proxy = PxProxy(self.config)
        status, headers, body = px_proxy.send_reverse_client_request(config=self.config, ctx=context)
        self.assertEqual(content, body.decode("utf-8"))

    @requests_mock.mock()
    def test_send_reverse_captcha_request(self, mock):
        content = 'captcha js content'
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/captcha/captcha.js', query_string='a=c&amp;u=cfe74220-f484-11e8-9b14-d7280325a290&amp;v=0701bb80-f482-11e8-8a31-a37cf9620569&amp;m=0')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        headers = {'host': px_constants.CAPTCHA_HOST,
                   px_constants.FIRST_PARTY_HEADER: '1',
                   px_constants.ENFORCER_TRUE_IP_HEADER: context.ip,
                   px_constants.FIRST_PARTY_FORWARDED_FOR: '127.0.0.1'}
        mock.get(
            url='https://captcha.px-cdn.net/PXfake_app_id/captcha.js?a=c&amp;u=cfe74220-f484-11e8-9b14-d7280325a290&amp;v=0701bb80-f482-11e8-8a31-a37cf9620569&amp;m=0',
            text=content, request_headers=headers, status_code=200, reason='OK')
        px_proxy = PxProxy(self.config)
        status, headers, body = px_proxy.send_reverse_captcha_request(config=self.config, ctx=context)
        self.assertEqual(content, body.decode("utf-8"))

    @requests_mock.mock()
    def test_send_reverse_xhr_request(self, mock):
        content = 'xhr content'
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/xhr/api/v1/collector', method='POST')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)

        headers = {'host': self.config.collector_host,
                   px_constants.FIRST_PARTY_HEADER: '1',
                   px_constants.ENFORCER_TRUE_IP_HEADER: context.ip,
                   px_constants.FIRST_PARTY_FORWARDED_FOR: '127.0.0.1'}
        mock.post(url='https://collector-pxfake_app_id.perimeterx.net/api/v1/collector', text=content,
                  request_headers=headers, status_code=200, reason='OK')
        px_proxy = PxProxy(self.config)
        status, headers, body = px_proxy.send_reverse_xhr_request(config=self.config, ctx=context, body=content)
        self.assertEqual(content, body.decode("utf-8"))



import json
import socket
import sys
import threading
import time
import traceback

from perimeterx import px_constants
from perimeterx import px_httpc
from perimeterx import px_utils

ACTIVITIES_BUFFER = []
CONFIG = {}


def init_activities_configuration(config):
    global CONFIG
    CONFIG = config

def _send_activities_chunk():
    global ACTIVITIES_BUFFER
    default_headers = {
        'Authorization': 'Bearer ' + CONFIG.auth_token,
        'Content-Type': 'application/json'
    }
    full_url = CONFIG.server_host + px_constants.API_ACTIVITIES
    chunk = ACTIVITIES_BUFFER[:CONFIG.max_buffer_len]
    for _ in range(len(chunk)):
        ACTIVITIES_BUFFER.pop(0)
    px_httpc.send(full_url=full_url, body=json.dumps(chunk), headers=default_headers, config=CONFIG, method='POST')

def send_activities_in_thread():
    if len(ACTIVITIES_BUFFER) >= CONFIG.max_buffer_len:
        CONFIG.logger.debug('Posting {} Activities'.format(len(ACTIVITIES_BUFFER)))
        t1 = threading.Thread(target=_send_activities_chunk)
        t1.daemon = True
        t1.start()
    else:
        CONFIG.logger.debug('NOT Posting {} Activities: '.format(len(ACTIVITIES_BUFFER)))

def send_to_perimeterx(activity_type, ctx, config, detail):
    try:
        if activity_type == 'page_requested' and not config.send_page_activities:
            print ('Page activities disabled in config - skipping.')
            return
        _details = {
            'http_method': ctx.http_method,
            'http_version': ctx.http_version,
            'module_version': config.module_version,
        }

        if len(detail.keys()) > 0:
            _details = dict(list(_details.items()) + list(detail.items()))

        data = {
            'type': activity_type,
            'headers': dict(ctx.headers),
            'timestamp': int(round(time.time() * 1000)),
            'socket_ip': ctx.ip,
            'px_app_id': config.app_id,
            'url': ctx.full_url,
            'details': _details,
            'vid': ctx.vid,
            'uuid': ctx.uuid
        }
        if activity_type == 'page_requested' or activity_type == 'block':
            px_utils.prepare_custom_params(config, _details)
            data['pxhd'] = ctx.pxhd

        ACTIVITIES_BUFFER.append(data)
    except:
        print (traceback.format_exception(*sys.exc_info()))
        return


def send_block_activity(ctx, config):
    send_to_perimeterx(px_constants.BLOCK_ACTIVITY, ctx, config, {
        'block_score': ctx.score,
        'block_uuid': ctx.uuid,
        'block_reason': ctx.block_reason,
        'http_version': ctx.http_version,
        'px_cookie': ctx.decoded_cookie,
        'risk_rtt': ctx.risk_rtt,
        'cookie_origin': ctx.cookie_origin,
        'block_action': ctx.block_action,
        'simulated_block': config.module_mode is px_constants.MODULE_MODE_MONITORING or ctx.monitored_route,
    })


def send_page_requested_activity(ctx, config):
    details = {
        'client_uuid': ctx.uuid,
        'pass_reason': ctx.pass_reason,
        'risk_rtt': ctx.risk_rtt
    }

    if ctx.decoded_cookie:
        details['px_cookie'] = ctx.decoded_cookie
    send_to_perimeterx(px_constants.PAGE_REQUESTED_ACTIVITY, ctx, config, details)


def send_enforcer_telemetry_activity(config, update_reason):
    details = {
        'enforcer_configs': config.telemetry_config,
        'node_name': socket.gethostname(),
        'os_name': sys.platform,
        'update_reason': update_reason,
        'module_version': config.module_version
    }
    body = {
        'type': px_constants.TELEMETRY_ACTIVITY,
        'timestamp': time.time(),
        'px_app_id': config.app_id,
        'details': details
    }

    headers = {
        'Authorization': 'Bearer ' + config.auth_token,
        'Content-Type': 'application/json'
    }
    config.logger.debug('Sending telemetry activity to PerimeterX servers')
    px_httpc.send(full_url=config.server_host + px_constants.API_ENFORCER_TELEMETRY, body=json.dumps(body),
                  headers=headers, config=config, method='POST')



import hashlib
import hmac
import base64
from Crypto.Cipher import AES


def create_hmac(str_to_hmac, config):
    try:
        return hmac.new(config.cookie_key.encode(), str_to_hmac.encode(), hashlib.sha256).hexdigest()
    except Exception:
        config.logger.debug("Failed to calculate hmac")
        return False

def decrypt_cookie(config, raw_cookie):
    """
    Decrypting the PerimeterX risk cookie using AES
    :return: Returns decrypted value if valid and False if not
    :rtype: Bool|String
    """
    logger = config.logger
    logger.debug("PxCookie[decrypt_cookie]")
    try:
        parts = raw_cookie.split(':', 3)
        if len(parts) != 3:
            return False
        salt = base64.b64decode(parts[0])
        iterations = int(parts[1])
        if iterations < 1 or iterations > 10000:
            return False
        data = base64.b64decode(parts[2])
        dk = hashlib.pbkdf2_hmac(hash_name='sha256', password=config.cookie_key.encode(), salt=salt, iterations=iterations, dklen=48)
        key = dk[:32]
        iv = dk[32:]
        cipher = AES.new(key, AES.MODE_CBC, iv)
        unpad = lambda s: s[0:-s[-1]]
        plaintext = unpad(cipher.decrypt(data))
        logger.debug("PxCookie[decrypt_cookie] cookie decrypted")
        return plaintext
    except Exception as e:
        logger.error('Encryption tool encountered a problem during the decryption process: ' + str(e))
        return False


def decode_cookie(config, raw_cookie):
    config.logger.debug("PxCookie[decode_cookie]")
    return base64.b64decode(raw_cookie)



import re
import traceback

from perimeterx import px_original_token_validator
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext
from perimeterx.px_cookie import PxCookie

mobile_error_pattern = re.compile("^\d+$")


def verify(ctx, config):
    """
    main verification function, verifying the content of the perimeterx risk cookie if exists
    :param PxContext ctx: perimeterx request context object
    :param PxConfig config: global configurations
    :return Bool: Returns True if verification succeeded and False if not
    """
    logger = config.logger
    try:
        if not ctx.px_cookies.keys():
            logger.debug('Cookie is missing')
            ctx.s2s_call_reason = 'no_cookie'
            return False

        if not config.cookie_key:
            logger.debug('No cookie key found, stopping cookie evaluation')
            ctx['s2s_call_reason'] = 'no_cookie_key'
            return False

        px_cookie_builder = PxCookie(config)

        cookie_version, px_cookie = px_cookie_builder.build_px_cookie(px_cookies=ctx.px_cookies,
                                                      user_agent=ctx.user_agent)
        # Mobile SDK traffic
        if px_cookie and ctx.is_mobile:
            if re.match(mobile_error_pattern, px_cookie._raw_cookie):
                logger.debug('Mobile special token - {}'.format(px_cookie.raw_cookie))
                ctx.s2s_call_reason = "mobile_error_" + px_cookie.raw_cookie
                if ctx.original_token is not None:
                    px_original_token_validator.verify(ctx, config)

                return False
            else:
                logger.debug('Mobile special token - no token')

        if not px_cookie.deserialize():
            cookie = px_cookie._hmac + ":" + px_cookie._raw_cookie
            logger.debug('Cookie decryption failed, value: {}'.format(cookie))
            ctx.px_cookie_raw = cookie_version + "=" + cookie
            ctx.s2s_call_reason = 'cookie_decryption_failed'
            return False

        ctx.score = px_cookie.get_score()
        ctx.uuid = px_cookie.get_uuid()
        if px_cookie.get_vid():
            ctx.vid = px_cookie.get_vid()
            ctx.enforcer_vid_source = 'risk_cookie'
        ctx.decoded_cookie = px_cookie.decoded_cookie
        ctx.cookie_hmac = px_cookie.get_hmac()
        ctx.block_action = px_cookie.get_action()

        if px_cookie.is_high_score():
            ctx.block_reason = 'cookie_high_score'
            logger.debug('Cookie with high score: {}'.format(ctx.score))
            return True

        if px_cookie.is_cookie_expired():
            ctx.s2s_call_reason = 'cookie_expired'
            msg = 'Cookie TTL is expired, value: {}, age: {}'
            logger.debug(msg.format(px_cookie.decoded_cookie, px_cookie.get_age()))
            return False

        if not px_cookie.is_secured():
            msg = 'Cookie HMAC validation failed, value: {}, user-agent: {}'
            logger.debug(msg.format(px_cookie.decoded_cookie, px_cookie._user_agent))
            ctx.s2s_call_reason = 'cookie_validation_failed'
            return False

        if ctx.sensitive_route:
            logger.debug('Sensitive route match, sending Risk API. path: {}'.format(ctx.uri))
            ctx.s2s_call_reason = 'sensitive_route'
            return False

        logger.debug('Cookie evaluation ended successfully, risk score: {}'.format(ctx.score))
        return True
    except Exception as err:
        traceback.print_exc()
        logger.error('Unexpected exception while evaluating Risk cookie. Error: {}'.format(err))
        ctx.px_cookie_raw = px_cookie._raw_cookie
        ctx.s2s_call_reason = 'cookie_decryption_failed'
        return False



from perimeterx.px_cookie import PxCookie


class PxCookieV3(PxCookie):

    def __init__(self, config, cookie, user_agent):
        super(PxCookieV3, self).__init__(config)
        self._config = config
        self._logger = config.logger
        self._user_agent = user_agent
        spliced_cookie = cookie.split(':')
        if len(spliced_cookie) is 4:
            self._hmac = spliced_cookie[0]
            self._raw_cookie = ':'.join(spliced_cookie[1:])
        else:
            self._raw_cookie = cookie

    def get_score(self):
        return self.decoded_cookie['s']

    def get_hmac(self):
        return self._hmac

    def get_action(self):
        return self.decoded_cookie['a']

    def is_cookie_format_valid(self):
        c = self.decoded_cookie
        return 't' in c and 'v' in c and 'u' in c and 's' in c and 'a' in c

    def is_secured(self):
        user_agent = self._user_agent
        str_hmac = self._raw_cookie + user_agent
        return self.is_cookie_valid(str_hmac)



import time

import requests
import os

#pylint: disable=import-error
if os.environ.get('SERVER_SOFTWARE','').startswith('Google'):
    import requests_toolbelt.adapters.appengine
    requests_toolbelt.adapters.appengine.monkeypatch()
#pylint: enable=import-error

def send(full_url, body, headers, config, method, raise_error = False):
    """
    Returns the appropriate response parameters according to blocking parameters
    :param string full_url:
    :param string body:
    :param dict headers:
    :param PxConfig config:
    :param string method:
    :return response: requests.response type
    """
    logger = config.logger
    try:
        start = time.time()
        if method == 'GET':
            response = requests.get(url=normalize_url(full_url), headers=headers, timeout=config.api_timeout, stream=True)
        else:
            response = requests.post(url=normalize_url(full_url), headers=headers, data=body, timeout=config.api_timeout)

        if response.status_code >= 400:
            logger.debug('PerimeterX server call failed: ' + str(response.status_code))
            return False
        finish = time.time()
        request_time = finish - start
        logger.debug('PerimeterX server call took {} ms'.format(request_time * 1000))
        return response
    except requests.exceptions.RequestException as err:
        logger.debug('PerimeterX Received Request Exception. Error: {}'.format(err))
        if raise_error:
            raise err


def normalize_url(url):
    if url.startswith("http://") or url.startswith("https://"):
        return url
    else:
        return "https://" + url



from perimeterx import px_constants
from perimeterx.px_testing_mode_handler import testing_mode_handling
from perimeterx.px_logger import Logger
import os

class PxConfig(object):
    def __init__(self, config_dict):
        app_id = config_dict.get('app_id')
        debug_mode = config_dict.get('debug_mode', False)
        module_mode = config_dict.get('module_mode', px_constants.MODULE_MODE_MONITORING)
        custom_logo = config_dict.get('custom_logo', None)
        testing_mode = config_dict.get('testing_mode', False)
        px_backend_host = config_dict.get('px_backend_url', None)
        max_buffer_len = config_dict.get('max_buffer_len', 30)
        self._px_app_id = app_id
        self._blocking_score = config_dict.get('blocking_score', 100)
        self._debug_mode = debug_mode
        self._module_version = config_dict.get('module_version', px_constants.MODULE_VERSION)
        self._module_version = px_constants.MODULE_VERSION.format(' GAE') if os.environ.get('SERVER_SOFTWARE','').startswith('Google') else px_constants.MODULE_VERSION.format('')
        self._module_mode = module_mode
        self._server_host = px_backend_host if px_backend_host else 'sapi.perimeterx.net' if app_id is None else px_constants.SERVER_URL.format(app_id.lower())
        self._collector_host = px_backend_host if px_backend_host else 'collector.perimeterx.net' if app_id is None else px_constants.COLLECTOR_URL.format(app_id.lower())
        self._encryption_enabled = config_dict.get('encryption_enabled', True)
        self._sensitive_headers = [*map(lambda header: header.lower(),
                                      config_dict.get('sensitive_headers', ['cookie', 'cookies']))]
        self._send_page_activities = config_dict.get('send_page_activities', True)
        self._api_timeout_ms = config_dict.get('api_timeout', 1000)
        self._custom_logo = custom_logo
        self._css_ref = config_dict.get('css_ref', '')
        self._js_ref = config_dict.get('js_ref', '')
        self._is_mobile = config_dict.get('is_mobile', False)
        self._monitor_mode = 0 if module_mode is px_constants.MODULE_MODE_MONITORING else 1
        self._module_enabled = config_dict.get('module_enabled', True)
        self._auth_token = config_dict.get('auth_token', None)
        self._is_mobile = config_dict.get('is_mobile', False)
        self._first_party = config_dict.get('first_party', True)
        self._first_party_xhr_enabled = config_dict.get('first_party_xhr_enabled', True)
        self._ip_headers = config_dict.get('ip_headers', [])
        self._proxy_url = config_dict.get('proxy_url', None)
        self._max_buffer_len = max_buffer_len if max_buffer_len > 0 else 1
        self._bypass_monitor_header = config_dict.get('bypass_monitor_header','')

        sensitive_routes = config_dict.get('sensitive_routes', [])
        if not isinstance(sensitive_routes, list):
            raise TypeError('sensitive_routes must be a list')
        self._sensitive_routes = sensitive_routes

        whitelist_routes = config_dict.get('whitelist_routes', [])
        if not isinstance(whitelist_routes, list):
            raise TypeError('whitelist_routes must be a list')
        self._whitelist_routes = whitelist_routes

        enforced_routes = config_dict.get('enforced_specific_routes', [])
        if not isinstance(enforced_routes, list):
            raise TypeError('enforced_specific_routes must be a list')
        self._enforced_specific_routes = enforced_routes

        monitored_routes = config_dict.get('monitored_specific_routes', [])
        if not isinstance(monitored_routes, list):
            raise TypeError('monitored_specific_routes must be a list')
        self._monitored_specific_routes = monitored_routes

        sensitive_routes_regex = config_dict.get('sensitive_routes_regex', [])
        if not isinstance(sensitive_routes_regex, list):
            raise TypeError('sensitive_routes_regex must be a list')
        self._sensitive_routes_regex = sensitive_routes_regex

        whitelist_routes_regex = config_dict.get('whitelist_routes_regex', [])
        if not isinstance(whitelist_routes_regex, list):
            raise TypeError('whitelist_routes_regex must be a list')
        self._whitelist_routes_regex = whitelist_routes_regex

        enforced_routes_regex = config_dict.get('enforced_specific_routes_regex', [])
        if not isinstance(enforced_routes_regex, list):
            raise TypeError('enforced_specific_routes must be a list')
        self._enforced_specific_routes_regex = enforced_routes_regex

        monitored_routes_regex = config_dict.get('monitored_specific_routes_regex', [])
        if not isinstance(monitored_routes_regex, list):
            raise TypeError('monitored_specific_routes_regex must be a list')
        self._monitored_specific_routes_regex = monitored_routes_regex

        self._block_html = 'BLOCK'
        self._telemetry_config = self.__create_telemetry_config()
        self._testing_mode = testing_mode
        self._auth_token = config_dict.get('auth_token', None)
        self._cookie_key = config_dict.get('cookie_key', None)
        self.__instantiate_user_defined_handlers(config_dict)
        self._logger = Logger(debug_mode, app_id)
        if testing_mode:
            self._custom_request_handler = testing_mode_handling

    @property
    def module_mode(self):
        return self._module_mode

    @property
    def app_id(self):
        return self._px_app_id

    @property
    def logger(self):
        return self._logger

    @property
    def auth_token(self):
        return self._auth_token

    @property
    def cookie_key(self):
        return self._cookie_key

    @property
    def server_host(self):
        return self._server_host

    @property
    def api_timeout(self):
        return self._api_timeout_ms / 1000.000

    @property
    def module_enabled(self):
        return self._module_enabled

    @property
    def ip_headers(self):
        return self._ip_headers

    @property
    def sensitive_headers(self):
        return self._sensitive_headers

    @property
    def proxy_url(self):
        return self._proxy_url

    @property
    def custom_request_handler(self):
        return self._custom_request_handler

    @property
    def blocking_score(self):
        return self._blocking_score

    @property
    def encryption_enabled(self):
        return self._encryption_enabled

    @property
    def module_version(self):
        return self._module_version

    @property
    def send_page_activities(self):
        return self._send_page_activities

    @property
    def custom_logo(self):
        return self._custom_logo

    @property
    def css_ref(self):
        return self._css_ref

    @property
    def js_ref(self):
        return self._js_ref

    @property
    def first_party(self):
        return self._first_party

    @property
    def first_party_xhr_enabled(self):
        return self._first_party_xhr_enabled

    @property
    def collector_host(self):
        return self._collector_host

    @property
    def get_user_ip(self):
        return self._get_user_ip

    @property
    def sensitive_routes(self):
        return self._sensitive_routes

    @property
    def whitelist_routes(self):
        return self._whitelist_routes

    @property
    def sensitive_routes_regex(self):
        return self._sensitive_routes_regex

    @property
    def whitelist_routes_regex(self):
        return self._whitelist_routes_regex

    @property
    def block_html(self):
        return self._block_html

    @property
    def additional_activity_handler(self):
        return self._additional_activity_handler

    @property
    def debug_mode(self):
        return self._debug_mode

    @property
    def max_buffer_len(self):
        return self._max_buffer_len

    @property
    def telemetry_config(self):
        return self._telemetry_config

    @property
    def enrich_custom_parameters(self):
        return self._enrich_custom_parameters

    @property
    def testing_mode(self):
        return self._testing_mode

    @property
    def bypass_monitor_header(self):
        return self._bypass_monitor_header

    @property
    def enforced_specific_routes(self):
        return self._enforced_specific_routes

    @property
    def monitored_specific_routes(self):
        return self._monitored_specific_routes

    @property
    def enforced_specific_routes_regex(self):
        return self._enforced_specific_routes_regex

    @property
    def monitored_specific_routes_regex(self):
        return self._monitored_specific_routes_regex

    def __instantiate_user_defined_handlers(self, config_dict):
        self._custom_request_handler = self.__set_handler('custom_request_handler', config_dict)
        self._get_user_ip = self.__set_handler('get_user_ip', config_dict)
        self._additional_activity_handler = self.__set_handler('additional_activity_handler', config_dict)
        self._enrich_custom_parameters = self.__set_handler('enrich_custom_parameters', config_dict)

    def __set_handler(self, function_name, config_dict):
        return config_dict.get(function_name) if config_dict.get(function_name) and callable(
            config_dict.get(function_name)) else None

    def __create_telemetry_config(self):
        config = self.__dict__
        mutated_config = {}
        for key, value in config.items():
            mutated_config[key[1:].upper()] = value
        return mutated_config




import unittest

from perimeterx.px_config import PxConfig
from perimeterx.px_cookie import PxCookie


class TestPXCookie(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        config = PxConfig({'app_id': 'fake_app_id'})
        cls.px_cookie = PxCookie(config)
        cls.config = config

    def test_build_cookie(self):
        px_cookies = {'_px3':'bd078865fa9627f626d6f7d6828ab595028d2c0974065ab6f6c5a9f80c4593cd:OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA='}
        cookie_version, cookie = self.px_cookie.build_px_cookie(px_cookies=px_cookies, user_agent='')
        self.assertEqual('bd078865fa9627f626d6f7d6828ab595028d2c0974065ab6f6c5a9f80c4593cd', cookie._hmac)
        self.assertEqual('OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA=', cookie._raw_cookie)

        px_cookies = {'_px3':'OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA='}
        cookie_version, cookie = self.px_cookie.build_px_cookie(px_cookies=px_cookies, user_agent='')
        self.assertEqual(cookie._hmac, '')


    def test_build_token(self):
        px_cookies = {'3':'bd078865fa9627f626d6f7d6828ab595028d2c0974065ab6f6c5a9f80c4593cd:OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA='}
        cookie_version, cookie = self.px_cookie.build_px_cookie(px_cookies=px_cookies, user_agent='')
        self.assertEqual('bd078865fa9627f626d6f7d6828ab595028d2c0974065ab6f6c5a9f80c4593cd', cookie._hmac)
        self.assertEqual('OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA=', cookie._raw_cookie)

        px_cookies = {'3':'OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA='}
        cookie_version, cookie = self.px_cookie.build_px_cookie(px_cookies=px_cookies, user_agent='')
        self.assertEqual(cookie._hmac, '')



import unittest

from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import Request

from perimeterx import px_cookie_validator
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext


class Test_PXCookieValidator(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cookie_key = 'Pyth0nS3crE7K3Y'
        cls.config = PxConfig({'app_id': 'app_id',
                           'cookie_key': cls.cookie_key})
        cls.headers = {'X-FORWARDED-FOR': '127.0.0.1',
                    'remote-addr': '127.0.0.1',
                   'content_length': '100',
                   'path': '/fake_app_id/init.js'}

    def test_verify_no_cookie(self):
        config = self.config

        builder = EnvironBuilder(headers= self.headers)

        env = builder.get_environ()
        request = Request(env)
        ctx = PxContext(request, PxConfig({'app_id': 'fake_app_id'}))
        verified = px_cookie_validator.verify(ctx, config)
        self.assertFalse(verified)
        self.assertEqual('no_cookie', ctx.s2s_call_reason)

    def test_verify_valid_cookie(self):
        config = self.config
        self.headers['cookie'] = '_px3=bd078865fa9627f626d6f7d6828ab595028d2c0974065ab6f6c5a9f80c4593cd:OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA='
        builder = EnvironBuilder(headers=self.headers)

        env = builder.get_environ()
        request = Request(env)
        ctx = PxContext(request, PxConfig({'app_id': 'fake_app_id'}))
        verified = px_cookie_validator.verify(ctx, config)
        self.assertTrue(verified)
        self.assertEqual('none', ctx.s2s_call_reason)
        del self.headers['cookie']

    def test_verify_decryption_failed(self):
        config = self.config
        cookie_value = '774958bcc233ea1a876b92ababf47086d8a4d95165bbd6f98b55d7e61afd2a05:ow3Er5dskpt8ZZ11CRiDMAueEi3ozJTqMBnYzsSM7/8vHTDA0so6ekhruiTrXa/taZINotR5PnTo78D5zM2pWw==:1000:uQ3Tdt7D3mSO5CuHDis3GgrnkGMC+XAghbHuNOE9x4H57RAmtxkTcNQ1DaqL8rx79bHl0iPVYlOcRmRgDiBCUoizBdUCjsSIplofPBLIl8WpfHDDtpxPKzz9I2rUEbFgfhFjiTY3rPGob2PUvTsDXTfPUeHnzKqbNTO8z7H6irFnUE='
        self.headers['cookie'] = '_px3=' + cookie_value
        builder = EnvironBuilder(headers=self.headers)
        env = builder.get_environ()
        request = Request(env)
        ctx = PxContext(request, PxConfig({'app_id': 'fake_app_id'}))
        verified = px_cookie_validator.verify(ctx, config)
        self.assertFalse(verified)
        self.assertEqual('cookie_decryption_failed', ctx.s2s_call_reason)
        self.assertEqual('_px3=' + cookie_value, ctx.px_cookie_raw)
        del self.headers['cookie']

    def test_verify_cookie_high_score(self):
        config = self.config
        self.headers['cookie'] = '_px3=bf46ceff75278ae166f376cbf741a7639060581035dd4e93641892c905dd0d67:EGFGcwQ2rum7KRmQCeSXBAUt1+25mj2DFJYi7KJkEliF3cBspdXtD2X03Csv8N8B6S5Bte/4ccCcETkBNDVxTw==:1000:x9x+oI6BISFhlKEERpf8HpZD2zXBCW9lzVfuRURHaAnbaMnpii+XjPEd7a7EGGUSMch5ramy3y+KOxyuX3F+LbGYwvn3OJb+u40zU+ixT1w5N15QltX+nBMhC7izC1l8QtgMuG/f3Nts5ebnec9j2V7LS5Y1/5b73rd9s7AMnug='
        builder = EnvironBuilder(headers=self.headers)
        env = builder.get_environ()
        request = Request(env)
        ctx = PxContext(request, PxConfig({'app_id': 'fake_app_id'}))
        verified = px_cookie_validator.verify(ctx, config)
        self.assertTrue(verified)
        self.assertEqual('none', ctx.s2s_call_reason)
        del self.headers['cookie']

    def test_verify_hmac_validation(self):
        config = self.config
        cookie_value = '774958bcc232343ea1a876b92ababf47086d8a4d95165bbd6f98b55d7e61afd2a05:ow3Er5dskpt8ZZ11CRiDMAueEi3ozJTqMBnYzsSM7/8vHTDA0so6ekhruiTrXa/taZINotR5PnTo78D5zM2pWw==:1000:uQ3Tdt7D3mSO5CuHDis3GgrnkGMC+XAghbHuNOE9x4H57RAmtxkTcNQ1DaqL8rx79bHl0iPVYlOcRmRgDiBCUoizBdUCjsSIplofPBLIl8WpfHDDtpxPKzz9I2rUEbFFjiTY3rPGob2PUvTsDXTfPUeHnzKqbNTO8z7H6irFnUE='
        self.headers['cookie'] = '_px3=' + cookie_value
        builder = EnvironBuilder(headers=self.headers)
        env = builder.get_environ()
        request = Request(env)
        ctx = PxContext(request, PxConfig({'app_id': 'fake_app_id'}))
        verified = px_cookie_validator.verify(ctx, config)
        self.assertFalse(verified)
        self.assertEqual('cookie_validation_failed', ctx.s2s_call_reason)
        self.assertEqual('', ctx.px_cookie_raw)
        del self.headers['cookie']

    def test_verify_expired_cookie(self):
        config = self.config
        cookie_value = '0d67bdf4a58c524b55b9cf0f703e4f0f3cbe23a10bd2671530d3c7e0cfa509eb:HOiYSw11ICB2A+HYx+C+l5Naxcl7hMeEo67QNghCQByyHlhWZT571ZKfqV98JFWg7TvbV9QtlrQtXakPYeIEjQ==:1000:+kuXS/iJUoEqrm8Fo4K0cTebsc4YQZu+f5bRGX0lC1T+l0g1gzRUuKiCtWTar28Y0wjch1ZQvkNy523Pxr07agVi/RL0SUktmEl59qGor+m4FLewZBVdcgx/Ya9kU0riis98AAR0zdTpTtoN5wpNbmztIpOZ0YejeD0Esk3vagU='
        self.headers['cookie'] = '_px3=' + cookie_value
        builder = EnvironBuilder(headers=self.headers)
        env = builder.get_environ()
        request = Request(env)
        ctx = PxContext(request, PxConfig({'app_id': 'fake_app_id'}))
        verified = px_cookie_validator.verify(ctx, config)
        self.assertFalse(verified)
        self.assertEqual('cookie_expired', ctx.s2s_call_reason)
        self.assertEqual('', ctx.px_cookie_raw)
        del self.headers['cookie']

    def test_cookie_decryption_failed_px_cookie_raw(self):
        config = self.config
        false_cookie = '_px3=774958bcc233ea1a876b92ababf47086d8a4d95165bbd6f98b55d7e61afd2a05:ow3Er5dskpt8ZZ11CRiDMAueEi3ozJTqMBnYzsSM7/8vHTDA0so6ekhruiTrXa/taZINotR5PnTo78D5zM2pWw==:1000:uQ3Tdt7D3mSO5CuHDis3GgrnkGMC+XAghbHuNOE9x4H57RAmtxkTcNQ1DaqL8rx79bHl0iPVYlOcRmRgDiBCUoizBdUCjsSIplofPBLIl8WpfHDDtpxPKzz9I2rUEbFgfhFjiTY3rPGob2PUvTsDXTfPUeHnzKqbNTO8z7H6irFnUE='
        self.headers['cookie'] = false_cookie
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/init.js')
        env = builder.get_environ()
        request = Request(env)
        ctx = PxContext(request, config)
        verified = px_cookie_validator.verify(ctx, config)
        self.assertEqual(ctx.px_cookie_raw, false_cookie)
        del self.headers['cookie']








PREFIX_PX_COOKIE_V1 = '_px'
PREFIX_PX_COOKIE_V3 = '_px3'
PREFIX_PX_TOKEN_V1 = '1'
PREFIX_PX_TOKEN_V3 = '3'
PREFIX_PX_DATA_ENRICHMENT = '_pxde'
PREFIX_PXVID = 'pxvid'
PREFIX_PXHD = '_pxhd'
MOBILE_SDK_HEADER = "x-px-authorization"
MOBILE_SDK_ORIGINAL_HEADER = "x-px-original-token"
MOBILE_SDK_CONNECTION_ERROR_CODE = '2'
MOBILE_SDK_PINNING_ERROR_CODE = '3'

BLOCK_TEMPLATE = 'block_template.mustache'
RATELIMIT_TEMPLATE = 'ratelimit.mustache'
CLIENT_HOST = 'client.perimeterx.net'
CAPTCHA_HOST = 'captcha.px-cdn.net'
COLLECTOR_URL = 'collector-{}.perimeterx.net'
SERVER_URL = 'sapi-{}.perimeterx.net'
CLIENT_FP_PATH = 'init.js'
CAPTCHA_FP_PATH = 'captcha'
XHR_FP_PATH = 'xhr'
MODULE_MODE_BLOCKING = 'active_blocking'
MODULE_MODE_MONITORING = 'monitor'
CLIENT_TP_PATH = 'main.min.js'
FIRST_PARTY_HEADER = 'x-px-first-party'
ENFORCER_TRUE_IP_HEADER = 'x-px-enforcer-true-ip'
EMPTY_GIF_B64 = 'R0lGODlhAQABAPAAAAAAAAAAACH5BAEAAAAALAAAAAABAAEAAAICRAEAOw=='
COLLECTOR_HOST = 'collector.perimeterx.net'
FIRST_PARTY_FORWARDED_FOR = 'X-FORWARDED-FOR'
MODULE_VERSION = 'Python 3 WSGI Module v1.2.0'
API_RISK = '/api/v3/risk'
PAGE_REQUESTED_ACTIVITY = 'page_requested'
BLOCK_ACTIVITY = 'block'
API_ENFORCER_TELEMETRY = '/api/v2/risk/telemetry'
API_ACTIVITIES = '/api/v1/collector/s2s'
TELEMETRY_ACTIVITY = 'enforcer_telemetry'
ACTION_CHALLENGE = 'j'
ACTION_BLOCK = 'b'
ACTION_RATELIMIT = 'r'
ACTION_CAPTCHA = 'c'
ALT_CAPTCHA_HOST = 'captcha.px-cloud.net'


# from perimeterx import px_activities_client
# from perimeterx.px_context import PxContext
# from perimeterx.px_config import PxConfig
# from perimeterx import px_constants
# import MagicMock
# import mock
# import unittest
#
#
# class TestPxActivitiesClient(unittest.TestCase):
#
#     @mock.patch('perimeterx.px_activities_client.send_to_perimeterx')
#     def test_send_to_perimeterx(self, mocked_event):
#         config = PxConfig({'app_id': 'fake_app_id'})
#         context = PxContext({}, config)
#         context.score = 100
#         context.uuid = 'uuid'
#         context.http_method = 's2s high score'
#         context.http_version = '1.1'
#         context.decoded_cookie = '{"u":"83c756b0-fbbd-11e8-956e-0d1c8297fb79","v":"75e52040-f6f9-11e8-bc4c-994724030299","t":1544368677086,"s":0,"a":"c"}'
#         context.risk_rtt = '2342'
#         context.cookie_origin = 'header'
#         context.block_action = 'c'
#         details = {
#             'block_score': context.score,
#             'client_uuid': context.uuid,
#             'block_reason': context.block_reason,
#             'http_method': context.http_method,
#             'http_version': context.http_version,
#             'px_cookie': context.decoded_cookie,
#             'risk_rtt': context.risk_rtt,
#             'cookie_origin': context.cookie_origin,
#             'block_action': context.block_action,
#             'module_version': px_constants.MODULE_VERSION,
#             'simulated_block': config.module_mode is px_constants.MODULE_MODE_MONITORING,
#         }
#         px_activities_client.send_to_perimeterx(activity_type='page_requested', ctx=context, config=config, detail=details)
#         print mock.call_args


import unittest
from perimeterx import px_original_token_validator
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext
from perimeterx import px_constants
from werkzeug.wrappers import Request
from werkzeug.test import EnvironBuilder


class TestPXOriginalTokenValidator(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.cookie_key = 'Pyth0nS3crE7K3Y'
        cls.config = PxConfig({'app_id': 'app_id',
                               'cookie_key': cls.cookie_key})
        cls.headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'content_length': '100',
                       'path': '/fake_app_id/init.js'}

    def test_verify(self):
        self.headers[
            px_constants.MOBILE_SDK_ORIGINAL_HEADER] = '3:bd078865fa9627f626d6f7d6828ab595028d2c0974065ab6f6c5a9f80c4593cd:OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA='
        self.headers[px_constants.MOBILE_SDK_HEADER] = '2'
        builder = EnvironBuilder(headers=self.headers)

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        verified = px_original_token_validator.verify(context, self.config)
        self.assertTrue(verified)
        self.assertEqual(context.vid, 'ce305f10-f17e-11e8-90f2-e7a14f96c498')
        self.assertEqual(context.decoded_original_token,
                         {'a': 'a', 's': 0, 'u': 'ce308620-f17e-11e8-90f2-e7a14f96c498', 't': 1663653730456,
                          'v': 'ce305f10-f17e-11e8-90f2-e7a14f96c498'})
        self.assertEqual(context.original_uuid, 'ce308620-f17e-11e8-90f2-e7a14f96c498')
        self.assertEqual(context.original_token_error, '')

    def test_decryption_error_token(self):
        self.headers[
            px_constants.MOBILE_SDK_ORIGINAL_HEADER] = '3:bd078865fa9627f626d6f7d6828ab595028d2c0974065ab6f6c5a9f80c4593cd:OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00asfafu4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA='
        self.headers[px_constants.MOBILE_SDK_HEADER] = '2'
        builder = EnvironBuilder(headers=self.headers)
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        verified = px_original_token_validator.verify(context, self.config)
        self.assertFalse(verified)
        self.assertEqual(context.original_token_error, 'decryption_failed')

    def test_validation_error_token(self):
        self.headers[
            px_constants.MOBILE_SDK_ORIGINAL_HEADER] = '3:bd078865fa9627f626d6f7d6828ab595028d2c0974ds065ab6f6c5afsaa9f80c4593cd:OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA='
        self.headers[px_constants.MOBILE_SDK_HEADER] = '2'
        builder = EnvironBuilder(headers=self.headers)
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        verified = px_original_token_validator.verify(context, self.config)
        self.assertFalse(verified)
        self.assertEqual(context.original_token_error, 'validation_failed')



from perimeterx.px_cookie import PxCookie


def verify(ctx, config):
    """
    main verification function, verifying the content of the perimeterx original token risk if exists
    :param pxContext ctx: perimeterx request context object
    :param PxConfig config: global configurations
    :return bool: Returns True if verification succeeded and False if not
    """
    logger = config.logger
    try:
        logger.debug('Original token found, Evaluating')
        original_token = ctx.original_token
        version, no_version_token = original_token.split(':', 1)
        px_cookie_builder = PxCookie(config)
        cookie_version, px_cookie = px_cookie_builder.build_px_cookie({version: no_version_token}, '')

        if not px_cookie.deserialize():
            logger.debug('Original token decryption failed, value: {}'.format(px_cookie.raw_cookie))
            ctx.original_token_error = 'decryption_failed'
            return False

        ctx.decoded_original_token = px_cookie.decoded_cookie
        ctx.vid = px_cookie.get_vid()
        ctx.original_uuid = px_cookie.get_uuid()
        if not px_cookie.is_secured():
            logger.debug('Original token HMAC validation failed, value: {}'.format(px_cookie.decoded_cookie))
            ctx.original_token_error = 'validation_failed'
            return False
        return True

    except Exception as e:
        logger.debug('Could not decrypt original token, exception was thrown, decryption failed. Error: {}'.format(e))
        ctx.original_token_error = 'decryption_failed'
        return False



import base64
import binascii
import hashlib
import json
import struct
import sys
import traceback
import hashlib
from time import time

from Crypto.Cipher import AES

from perimeterx import px_enc_utils
from perimeterx.px_constants import *


class PxCookie(object):

    def __init__(self, config):
        self._config = config
        self._logger = config.logger
        self._raw_cookie = ''
        self._hmac = ''

    def build_px_cookie(self, px_cookies, user_agent=''):
        self._logger.debug("PxCookie[build_px_cookie]")
        if not px_cookies:
            self._logger.debug('Cookie is missing')
            return None

        px_cookie_keys = px_cookies.keys()
        sorted(px_cookie_keys, reverse=True)
        for prefix in px_cookie_keys:
            if prefix == PREFIX_PX_TOKEN_V1 or prefix == PREFIX_PX_COOKIE_V1:
                self._logger.debug('Cookie/Token V1 found, evaluating..')
                from perimeterx.px_cookie_v1 import PxCookieV1
                return prefix, PxCookieV1(self._config, px_cookies[prefix])
            if prefix == PREFIX_PX_TOKEN_V3 or prefix == PREFIX_PX_COOKIE_V3:
                self._logger.debug('Cookie/Token V3 found, evaluating..')
                from perimeterx.px_cookie_v3 import PxCookieV3
                ua = ''
                if prefix == PREFIX_PX_COOKIE_V3:
                    ua = user_agent
                return prefix, PxCookieV3(self._config, px_cookies[prefix], ua)

        self._logger.debug('Cookie is missing')

    def decode_cookie(self):
        self._logger.debug("PxCookie[decode_cookie]")
        return px_enc_utils.decode_cookie(self._config, self._raw_cookie)

    def decrypt_cookie(self):
        """
        Decrypting the PerimeterX risk cookie using AES
        :return: Returns decrypted value if valid and False if not
        :rtype: Bool|String
        """
        try:
            parts = self._raw_cookie.split(':', 3)
            if len(parts) != 3:
                return False
            salt = base64.b64decode(parts[0])
            iterations = int(parts[1])
            if iterations < 1 or iterations > 10000:
                return False
            data = base64.b64decode(parts[2])
            dk = hashlib.pbkdf2_hmac(hash_name='sha256', password=self._config.cookie_key.encode(), salt=salt, iterations=iterations, dklen=48)
            key = dk[:32]
            iv = dk[32:]
            cipher = AES.new(key, AES.MODE_CBC, iv)
            unpad = lambda s: s[0:-s[-1]]
            plaintext = unpad(cipher.decrypt(data))
            self._logger.debug('cookie decrypted')
            return plaintext
        except:
            print (traceback.format_exception(*sys.exc_info()))
            return None

    def is_cookie_expired(self):
        """
        Checks if cookie validity time expired.
        :return: Returns True if valid and False if not
        :rtype: Bool
        """
        now = int(round(time() * 1000))
        expire = self.get_timestamp()
        return now > expire

    def is_cookie_valid(self, str_to_hmac):
        """
        Checks if cookie hmac signing match the request.
        :return: Returns True if valid and False if not
        :rtype: Bool
        """
        try:
            calculated_digest = px_enc_utils.create_hmac(str_to_hmac, self._config)
            return self.get_hmac() == calculated_digest
        except Exception as err:
            self._logger.debug("failed to calculate hmac: {}".format(err))
            return False

    def deserialize(self):
        logger = self._logger
        if self._config.encryption_enabled:
            cookie = px_enc_utils.decrypt_cookie(config=self._config, raw_cookie=self._raw_cookie)
        else:
            cookie = px_enc_utils.decode_cookie(config=self._config, raw_cookie=self._raw_cookie)

        if not cookie:
            return False

        logger.debug('Original token deserialized: {}'.format(cookie))
        self.decoded_cookie = json.loads(cookie)
        return self.is_cookie_format_valid()

    def is_high_score(self):
        return self.get_score() >= self._config.blocking_score

    def get_timestamp(self):
        return self.decoded_cookie['t']

    def get_uuid(self):
        return self.decoded_cookie['u']

    def get_vid(self):
        return self.decoded_cookie['v']

    def get_age(self):
        return int(round(time() * 1000)) - self.decoded_cookie['t']

    def get_hmac(self):
        return self._hmac



import unittest
from perimeterx import px_httpc
import requests_mock
from perimeterx.px_config import PxConfig


class TestPXHttpc(unittest.TestCase):
    def test_send(self):
        with requests_mock.mock() as m:
            config = PxConfig({'app_id': 'PXfake_app_id'})
            full_url = 'this_url.com/uri'
            method = 'POST'
            body = 'content to post'

            headers = {'content-type': 'application/json'}

            m.post('https://' + full_url)
            response = px_httpc.send(full_url=full_url, config=config,method=method,body=body,headers=headers)
            m.called
            m.get('https://' + full_url)
            method = 'GET'
            response = px_httpc.send(full_url=full_url, config=config,method=method,body=body,headers=headers)
            self.assertEqual(m.call_count, 2)


import json
import time
import os
import requests

#pylint: disable=import-error
if os.environ.get('SERVER_SOFTWARE','').startswith('Google'):
    import requests_toolbelt.adapters.appengine
    requests_toolbelt.adapters.appengine.monkeypatch()
#pylint: enable=import-error

from perimeterx import px_constants
from perimeterx import px_httpc
from perimeterx import px_utils

custom_params = {
    'custom_param1': '',
    'custom_param2': '',
    'custom_param3': '',
    'custom_param4': '',
    'custom_param5': '',
    'custom_param6': '',
    'custom_param7': '',
    'custom_param8': '',
    'custom_param9': '',
    'custom_param10': ''
}


def send_risk_request(ctx, config):
    """
    :param PxContext ctx:
    :param PxConfig config:
    :return dict:
    """
    start = time.time()
    body = prepare_risk_body(ctx, config)
    default_headers = {
        'Authorization': 'Bearer ' + config.auth_token,
        'Content-Type': 'application/json'
    }
    try:
        response = px_httpc.send(full_url=config.server_host + px_constants.API_RISK, body=json.dumps(body),
                                 config=config, headers=default_headers, method='POST', raise_error = True)
        if response:
            config.logger.debug('Risk response: {}'.format(response.content))
            return json.loads(response.content)
        return False
    except requests.exceptions.Timeout:
        ctx.pass_reason = 's2s_timeout'
        risk_rtt = time.time() - start
        config.logger.debug('Risk API timed out, round_trip_time: {}'.format(risk_rtt))
        return False
    except requests.exceptions.RequestException as e:
        ctx.pass_reason = 's2s_error'
        config.logger.debug('Unexpected exception in Risk API call: {}'.format(e))
        return False

def verify(ctx, config):
    """
    :param PxContext ctx:
    :param pxConfig config:
    :return bool: is request verified
    """

    logger = config.logger
    logger.debug('Evaluating Risk API request, call reason: {}'.format(ctx.s2s_call_reason))
    try:
        start = time.time()
        response = send_risk_request(ctx, config)
        risk_rtt = time.time() - start
        logger.debug('Risk call took {} ms'.format(risk_rtt))

        if response:
            ctx.score = response.get('score')
            ctx.uuid = response.get('uuid')
            ctx.block_action = response.get('action')
            ctx.risk_rtt = risk_rtt
            ctx.pxde = response.get('data_enrichment', {})
            ctx.pxde_verified = True
            response_pxhd = response.get('pxhd', '')
            #Do not set cookie if there's already a valid pxhd
            ctx.response_pxhd = response_pxhd
            if ctx.score >= config.blocking_score:
                if response.get('action') == px_constants.ACTION_CHALLENGE and \
                        response.get('action_data') is not None and \
                        response.get('action_data').get('body') is not None:

                    logger.debug("received javascript challenge action")
                    ctx.block_action_data = response.get('action_data').get('body')
                    ctx.block_reason = 'challenge'

                elif response.get('action') is px_constants.ACTION_RATELIMIT:
                    logger.debug("received javascript ratelimit action")
                    ctx.block_reason = 'exceeded_rate_limit'

                else:
                    logger.debug("block score threshold reached, will initiate blocking")
                    ctx.block_reason = 's2s_high_score'
            else:
                ctx.pass_reason = 's2s'

            msg = 'Risk API response returned successfully, risk score: {}, round_trip_time: {} ms'
            logger.debug(msg.format(ctx.score, risk_rtt))
            return True
        else:
            return False
    except Exception as err:
        logger.error('Risk API request failed. Error: {}'.format(err))
        return False


def prepare_risk_body(ctx, config):
    logger = config.logger
    risk_mode  = 'monitor' if config.module_mode == px_constants.MODULE_MODE_MONITORING or ctx.monitored_route else 'active_blocking'
    body = {
        'request': {
            'ip': ctx.ip,
            'headers': format_headers(ctx.headers),
            'uri': ctx.uri,
            'url': ctx.full_url,
            'firstParty': 'true' if config.first_party else 'false'
        },
        'additional': {
            's2s_call_reason': ctx.s2s_call_reason,
            'http_method': ctx.http_method,
            'http_version': ctx.http_version,
            'module_version': config.module_version,
            'risk_mode': risk_mode,
            'cookie_origin': ctx.cookie_origin
        }
    }
    if ctx.vid:
        body['vid'] = ctx.vid
        body['additional']['enforcer_vid_source'] = ctx.enforcer_vid_source
    if ctx.uuid:
        body['uuid'] = ctx.uuid
    if ctx.cookie_hmac:
        body['additional']['px_cookie_hmac'] = ctx.cookie_hmac
    if ctx.cookie_names:
        body['additional']['request_cookie_names'] = ctx.cookie_names
    if ctx.pxhd:
        body['pxhd'] = ctx.pxhd

    body = add_original_token_data(ctx, body)

    px_utils.prepare_custom_params(config, body['additional'])

    if ctx.s2s_call_reason == 'cookie_decryption_failed':
        logger.debug('attaching orig_cookie to request')
        body['additional']['px_cookie_raw'] = ctx.px_cookie_raw

    if ctx.s2s_call_reason in ['cookie_expired', 'cookie_validation_failed']:
        logger.debug('attaching px_cookie to request')
        body['additional']['px_cookie'] = ctx.decoded_cookie


    return body


def add_original_token_data(ctx, body):
    if ctx.original_uuid:
        body['additional']['original_uuid'] = ctx.original_uuid
    if ctx.original_token_error:
        body['additional']['original_token_error'] = ctx.original_token_error
    if ctx.original_token:
        body['additional']['original_token'] = ctx.original_token
    if ctx.decoded_original_token:
        body['additional']['decoded_original_token'] = ctx.decoded_original_token
    return body


def format_headers(headers):
    ret_val = []
    for key in headers.keys():
        ret_val.append({'name': key, 'value': headers[key]})
    return ret_val



import unittest

import mock
from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import Request, Response

import perimeterx.px_constants as px_constants
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext
from perimeterx.px_request_verifier import PxRequestVerifier


class TestPxRequestVerifier(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING
                               })
        cls.headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'content_length': '100'}
        cls.request_handler = PxRequestVerifier(cls.config)

    def test_verify_request_fp_client_passed(self):
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/init.js')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        response = Response("client data")
        with mock.patch('perimeterx.px_proxy.PxProxy.handle_reverse_request', return_value=response):
            response = self.request_handler.verify_request(context, request)
            self.assertEqual(response.data.decode("utf-8"), "client data")

    def test_verify_static_url(self):
        builder = EnvironBuilder(headers=self.headers, path='/fake.css')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        response = self.request_handler.verify_request(context, request)
        self.assertEqual(response, True)

    def test_verify_whitelist(self):
        config = PxConfig({'app_id': 'PXfake_app_id', 'whitelist_routes': ['whitelisted']})
        builder = EnvironBuilder(headers=self.headers, path='/whitelisted')
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, config)
        response = self.request_handler.verify_request(context, request)
        self.assertEqual(response, True)

    def test_handle_verification_pass(self):
        config = PxConfig({'app_id': 'PXfake_app_id', 'whitelist_routes': ['whitelisted']})
        builder = EnvironBuilder(headers=self.headers, path='/whitelisted')
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, config)
        context.score = 50
        response = self.request_handler.handle_verification(context, request)
        self.assertEqual(response, True)

    def test_handle_verification_failed(self):
        config = PxConfig({'app_id': 'PXfake_app_id', 'whitelist_routes': ['whitelisted']})
        builder = EnvironBuilder(headers=self.headers, path='/whitelisted')
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, config)
        context.score = 100
        response = self.request_handler.handle_verification(context, request)
        self.assertEqual(response.status, '403 Forbidden')

    def test_handle_monitor(self):
        config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_MONITORING
                               });
        request_handler = PxRequestVerifier(config)
        builder = EnvironBuilder(headers=self.headers, path='/')
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, request_handler.config)
        context.score = 100
        response = request_handler.handle_verification(context, request)
        self.assertEqual(response, True)

    def test_bypass_monitor_header_enabled(self):
        config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_MONITORING,
                               'bypass_monitor_header': 'x-px-block'
                               });
        headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'x-px-block': '1',
                       'content_length': '100'}
        request_handler = PxRequestVerifier(config)
        builder = EnvironBuilder(headers=headers, path='/')
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, request_handler.config)
        context.score = 100
        response = request_handler.handle_verification(context, request)
        self.assertEqual(response.status, '403 Forbidden')

    def test_bypass_monitor_header_disabled(self):
        config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_MONITORING,
                               'bypass_monitor_header': 'x-px-block'
                               });
        headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'x-px-block': '0',
                       'content_length': '100'}
        request_handler = PxRequestVerifier(config)
        builder = EnvironBuilder(headers=headers, path='/')
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, request_handler.config)
        context.score = 100
        response = request_handler.handle_verification(context, request)
        self.assertEqual(response, True)

    def test_bypass_monitor_header_configured_but_missing(self):
        config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_MONITORING,
                               'bypass_monitor_header': 'x-px-block'
                               });
        headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'content_length': '100'}
        request_handler = PxRequestVerifier(config)
        builder = EnvironBuilder(headers=headers, path='/')
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, request_handler.config)
        context.score = 100
        response = request_handler.handle_verification(context, request)
        self.assertEqual(response, True)

    def test_bypass_monitor_header_on_valid_request(self):
        config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_MONITORING,
                               'bypass_monitor_header': 'x-px-block'
                               });
        headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'x-px-block': '1',
                       'content_length': '100'}
        request_handler = PxRequestVerifier(config)
        builder = EnvironBuilder(headers=headers, path='/')
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, request_handler.config)
        context.score = 0
        response = request_handler.handle_verification(context, request)
        self.assertEqual(response, True)

    def test_specific_enforced_routes_with_enforced_route(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING,
                               'enforced_specific_routes': ['/profile'],
                               });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                     'remote-addr': '127.0.0.1',
                     'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/profile')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      context.score = 100
      response = request_handler.handle_verification(context, request)
      self.assertEqual(response.status, '403 Forbidden')

    def test_specific_enforced_routes_with_enforced_route_regex(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING,
                               'enforced_specific_routes_regex': [r'^/profile$'],
                               });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                     'remote-addr': '127.0.0.1',
                     'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/profile')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      context.score = 100
      response = request_handler.handle_verification(context, request)
      self.assertEqual(response.status, '403 Forbidden')

    def test_specific_enforced_routes_with_unenforced_route(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING,
                               'enforced_specific_routes': ['/profile'],
                               });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                     'remote-addr': '127.0.0.1',
                     'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      context.score = 100
      response = request_handler.verify_request(context, request)
      self.assertEqual(response, True)

      def test_specific_enforced_routes_with_unenforced_route_regex(self):
        config = PxConfig({'app_id': 'PXfake_app_id',
                                 'auth_token': '',
                                 'module_mode': px_constants.MODULE_MODE_BLOCKING,
                                 'enforced_specific_routes_regex': [r'^/profile$'],
                                 });
        headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'content_length': '100'}
        request_handler = PxRequestVerifier(config)
        builder = EnvironBuilder(headers=headers, path='/')
        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, request_handler.config)
        context.score = 100
        response = request_handler.verify_request(context, request)
        self.assertEqual(response, True)

    def test_monitor_specific_routes_in_blocking_mode(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING,
                               'monitored_specific_routes': ['/profile'],
                               });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                     'remote-addr': '127.0.0.1',
                     'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/profile')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      response = request_handler.verify_request(context, request)
      self.assertEqual(context.monitored_route, True)

    def test_monitor_specific_routes_in_blocking_mode_regex(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING,
                               'monitored_specific_routes_regex': [r'^/profile$'],
                               });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                     'remote-addr': '127.0.0.1',
                     'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/profile')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      response = request_handler.verify_request(context, request)
      self.assertEqual(context.monitored_route, True)

    def test_monitor_specific_routes_in_blocking_mode_should_block_other_routes(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING,
                               'monitored_specific_routes': ['/profile'],
                               });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                     'remote-addr': '127.0.0.1',
                     'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      context.score = 100
      response = request_handler.handle_verification(context, request)
      self.assertEqual(response.status, '403 Forbidden')

    def test_monitor_specific_routes_in_blocking_mode_should_block_other_routes_regex(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING,
                               'monitored_specific_routes_regex': [r'^/profile$'],
                               });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                     'remote-addr': '127.0.0.1',
                     'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/profile/me')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      context.score = 100
      response = request_handler.handle_verification(context, request)
      self.assertEqual(response.status, '403 Forbidden')

    def test_enforced_specific_routes_overrides_monitor_specific_routes(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                             'auth_token': '',
                             'module_mode': px_constants.MODULE_MODE_BLOCKING,
                             'monitored_specific_routes': ['/profile'],
                             'enforced_specific_routes': ['/profile', '/login'],
                             });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                   'remote-addr': '127.0.0.1',
                   'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/profile')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      context.score = 100
      response = request_handler.handle_verification(context, request)
      self.assertEqual(response.status, '403 Forbidden')

    def test_enforced_specific_routes_overrides_monitor_specific_routes(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                             'auth_token': '',
                             'module_mode': px_constants.MODULE_MODE_BLOCKING,
                             'monitored_specific_routes_regex': [r'^/profile$'],
                             'enforced_specific_routes_regex': [r'^/profile$', r'^/login$'],
                             });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                   'remote-addr': '127.0.0.1',
                   'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/profile')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      context.score = 100
      response = request_handler.handle_verification(context, request)
      self.assertEqual(response.status, '403 Forbidden')

    def test_monitor_specific_routes_with_enforced_specific_routes(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING,
                               'monitored_specific_routes': ['/profile'],
                                'enforced_specific_routes': ['/login'],
                               });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                     'remote-addr': '127.0.0.1',
                     'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/profile')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      response = request_handler.verify_request(context, request)
      self.assertEqual(context.monitored_route, True)

    def test_monitor_specific_routes_with_enforced_specific_routes_regex(self):
      config = PxConfig({'app_id': 'PXfake_app_id',
                               'auth_token': '',
                               'module_mode': px_constants.MODULE_MODE_BLOCKING,
                               'monitored_specific_routes_regex': [r'^/profile$'],
                                'enforced_specific_routes_regex': [r'^/login$'],
                               });
      headers = {'X-FORWARDED-FOR': '127.0.0.1',
                     'remote-addr': '127.0.0.1',
                     'content_length': '100'}
      request_handler = PxRequestVerifier(config)
      builder = EnvironBuilder(headers=headers, path='/profile')
      env = builder.get_environ()
      request = Request(env)
      context = PxContext(request, request_handler.config)
      response = request_handler.verify_request(context, request)
      self.assertEqual(context.monitored_route, True)


import base64
import json

import pystache

from perimeterx import px_constants
from perimeterx import px_template
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext


class PXBlocker(object):
    def __init__(self):
        self.mustache_renderer = pystache.Renderer()
        self.ratelimit_rendered_page = self.mustache_renderer.render(
            px_template.get_template(px_constants.RATELIMIT_TEMPLATE), {})

    def handle_blocking(self, ctx, config):
        """
        Returns the appropriate response parameters according to blocking parameters
        :param PxContext ctx:
        :param PxConfig config:
        :return (string, array, string): response string, headers array, and status string
        """

        logger = config.logger
        action = ctx.block_action
        status = '403 Forbidden'

        is_json_response = self.is_json_response(ctx)
        if is_json_response:
            content_type = 'application/json'
        else:
            content_type = 'text/html'
        headers = {'Content-Type': content_type}

        if action is px_constants.ACTION_CHALLENGE:
            logger.debug('Enforcing action: Challenge page is served')
            blocking_props = ctx.block_action_data
            blocking_response = blocking_props

        elif action is px_constants.ACTION_RATELIMIT:
            logger.debug('Enforcing action: Rate limit page is served')
            blocking_props = None
            blocking_response = self.ratelimit_rendered_page
            status = '429 Too Many Requests'

        else:  # block
            logger.debug('Enforcing action: Block page is served')
            blocking_props = self.prepare_properties(ctx, config)
            blocking_response = self.mustache_renderer.render(px_template.get_template(px_constants.BLOCK_TEMPLATE),
                                                              blocking_props)

        if ctx.is_mobile:
            page_response = json.dumps({
                'action': parse_action(ctx.block_action),
                'uuid': ctx.uuid,
                'vid': ctx.vid,
                'appId': config.app_id,
                'page': str(base64.b64encode(blocking_response.encode())),
                'collectorURL': 'https://' + config.collector_host
            })
            return page_response, headers, status

        if is_json_response:
            logger.debug('Serving advanced blocking response')
            blocking_response = json.dumps(blocking_props)

        blocking_response = str(blocking_response)
        return blocking_response, headers, status

    def prepare_properties(self, ctx, config):
        app_id = config.app_id
        vid = ctx.vid
        uuid = ctx.uuid
        custom_logo = config.custom_logo if config.custom_logo else ''
        is_mobile_num = 1 if ctx.is_mobile else 0
        captcha_uri = 'captcha.js?a={}&u={}&v={}&m={}'.format(ctx.block_action, uuid, vid, is_mobile_num)
        alt_captcha_src = '//{}/{}/{}'.format(px_constants.ALT_CAPTCHA_HOST, app_id, captcha_uri)

        if config.first_party and not ctx.is_mobile:
            prefix = app_id[2:]
            js_client_src = '/{}/{}'.format(prefix, px_constants.CLIENT_FP_PATH)
            captcha_src = '/{}/{}/{}'.format(prefix, px_constants.CAPTCHA_FP_PATH, captcha_uri)
            host_url = '/{}/{}'.format(prefix, px_constants.XHR_FP_PATH)
        else:
            js_client_src = '//{}/{}/main.min.js'.format(px_constants.CLIENT_HOST, app_id)
            captcha_src = '//{}/{}/{}'.format(px_constants.CAPTCHA_HOST, app_id, captcha_uri)
            host_url = px_constants.COLLECTOR_URL.format(app_id.lower())

        return {
            'appId': app_id,
            'vid': vid,
            'uuid': uuid,
            'customLogo': custom_logo,
            'cssRef': config.css_ref,
            'jsRef': config.js_ref,
            'hostUrl': host_url,
            'jsClientSrc': js_client_src,
            'firstPartyEnabled': 'true' if config.first_party else 'false',
            'blockScript': captcha_src,
            'altBlockScript': alt_captcha_src
        }

    def is_json_response(self, ctx):
        headers = ctx.headers
        if ctx.block_action is not px_constants.ACTION_RATELIMIT:
            for item in headers.keys():
                if item.lower() == 'accept' or item.lower() == 'content-type':
                    item_arr = headers[item].split(',')
                    for header_item in item_arr:
                        if header_item.strip() == 'application/json':
                            return True
        return False


def parse_action(action):
    if 'b' == action:
        return 'block'
    elif 'j' == action:
        return 'challege'
    elif 'r' == action:
        return 'ratelimit'
    else:
        return 'captcha'



from werkzeug.wrappers import Response

from perimeterx import px_activities_client
from perimeterx import px_api
from perimeterx import px_blocker
from perimeterx import px_constants
from perimeterx import px_cookie_validator
from perimeterx import px_utils
from perimeterx.px_proxy import PxProxy


class PxRequestVerifier(object):

    def __init__(self, config):
        self.config = config
        self.logger = config.logger
        self._PXBlocker = px_blocker.PXBlocker()

    def verify_request(self, ctx, request):
        uri = ctx.uri
        px_proxy = PxProxy(self.config)
        if px_proxy.should_reverse_request(uri):
            return px_proxy.handle_reverse_request(self.config, ctx, request.get_data())
        if px_utils.is_static_file(ctx):
            self.logger.debug('Filter static file request. uri: {}'.format(uri))
            return True
        if request.environ.get('REQUEST_METHOD') == 'OPTIONS':
            self.logger.debug('The request method was whitelisted, OPTIONS')
            return True
        if ctx.whitelist_route:
            self.logger.debug('The requested uri is whitelisted, passing request')
            return True
        if len(self.config.enforced_specific_routes) > 0 and not ctx.enforced_route:
            self.logger.debug('The request uri {} is not listed in specific routes to enforce, passing request.'.format(uri))
            return True
        # PX Cookie verification
        if not px_cookie_validator.verify(ctx, self.config):
            # Server-to-Server verification fallback
            if not px_api.verify(ctx, self.config):
                self.report_pass_traffic(ctx)
                return True
        return self.handle_verification(ctx, request)

    def handle_verification(self, ctx, request):
        config = self.config
        logger = config.logger
        score = ctx.score
        data = None
        headers = None
        status = None
        pass_request = False
        if score < config.blocking_score:
            logger.debug('Risk score is lower than blocking score')
            self.report_pass_traffic(ctx)
            pass_request = True
        else:
            logger.debug('Risk score is higher or equal than blocking score')
            self.report_block_traffic(ctx)
            should_bypass_monitor = config.bypass_monitor_header and ctx.headers.get(config.bypass_monitor_header) == '1';
            if config.additional_activity_handler:
                config.additional_activity_handler(ctx, config)
            if ctx.monitored_route or (config.module_mode == px_constants.MODULE_MODE_MONITORING and not should_bypass_monitor):
                return True
            else:
                data, headers, status = self.px_blocker.handle_blocking(ctx=ctx, config=config)
                response_function = generate_blocking_response(data, headers, status)

        if config.custom_request_handler:
            data, headers, status = config.custom_request_handler(ctx, self.config, request)
            if data and headers and status:
                return generate_blocking_response(data, headers, status)

        if pass_request:
            return True
        else:
            return response_function

    def report_pass_traffic(self, ctx):
        px_activities_client.send_page_requested_activity(ctx, self.config)

    def report_block_traffic(self, ctx):
        px_activities_client.send_block_activity(ctx, self.config)

    @property
    def px_blocker(self):
        return self._PXBlocker


def generate_blocking_response(data, headers, status):
    result = Response(data)
    if headers is not None:
        result.headers = headers
    if status is not None:
        result.status = status
    return result



import os


def get_path():
    return os.path.dirname(os.path.abspath(__file__))


def get_content(template):
    template_path = '{}/templates/{}'.format(get_path(), template)
    file_handler = open(template_path, "r")
    content = file_handler.read()
    return content


def get_template(template_name):
    return get_content(template_name)



import unittest

from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import Request

from perimeterx import px_constants
from perimeterx import px_utils
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext


class Test_PXUtils(unittest.TestCase):

    def test_merge_two_dicts(self):
        dict1 = {'a': '1'}
        dict2 = {'b': '2'}
        merged_dict = px_utils.merge_two_dicts(dict1, dict2)
        self.assertDictEqual(merged_dict, {'a': '1', 'b': '2'})

    def test_handle_proxy_headers(self):
        headers_sample = {'ddd': 'not_proxy_url', px_constants.FIRST_PARTY_FORWARDED_FOR: 'proxy_url'}
        headers_sample = px_utils.handle_proxy_headers(headers_sample, '127.0.0.1', False)
        self.assertEqual(headers_sample[px_constants.FIRST_PARTY_FORWARDED_FOR], '127.0.0.1')
        headers_sample = {'ddd': 'not_proxy_url'}
        headers_sample = px_utils.handle_proxy_headers(headers_sample, '127.0.0.1', False)
        self.assertEqual(headers_sample[px_constants.FIRST_PARTY_FORWARDED_FOR], '127.0.0.1')

    def test_is_static_file(self):
        config = PxConfig({'app_id' : 'fake_app_id'})
        headers = {'X-FORWARDED-FOR': '127.0.0.1',
                   'remote-addr': '127.0.0.1',
                   'content_length': '100',
                   'cookie': '_px3=bd078865fa9627f626d6f7d6828ab595028d2c0974065ab6f6c5a9f80c4593cd:OCIluokZHHvqrWyu8zrWSH8Vu7AefCjrd4CMx/NXsX58LzeV40EZIlPG4gsNMoAYzH88s/GoZwv+DpQa76C21A==:1000:zwT+Rht/YGDNWKkzHtJAB7IiI00u4fOePL/3xWMs1nZ93lzW1XvAMGR2hLlHBmOv8O0CpylEQOZZTK1uQMls6O28Y8aQnTo5DETLkrbhpwCVeNjOcf8GVKTckITwuHfXbEcfHbdtb68s1+jHv1+vt/w/6HZqTzanaIsvFVp8vmA='}
        builder = EnvironBuilder(headers=headers, path='/sample.css')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, config)
        self.assertTrue(px_utils.is_static_file(context))
        builder = EnvironBuilder(headers=headers, path='/sample.html')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, config)
        self.assertFalse(px_utils.is_static_file(context))

    def enrich_custom_parameters(self, params):
        params['custom_param1'] = '1'
        params['custom_param2'] = '2'
        params['custom'] = '6'
        params['custom_param10'] = '10'
        return params

    def test_prepare_risk_body(self):
        config = PxConfig({'app_id': 'app_id', 'enrich_custom_parameters': self.enrich_custom_parameters})
        additional = {}

        px_utils.prepare_custom_params(config, additional)
        self.assertEqual(additional.get('custom_param1'), '1')
        self.assertEqual(additional.get('custom_param2'), '2')
        self.assertEqual(additional.get('custom_param10'), '10')
        self.assertFalse(additional.get('custom_param11'))
        self.assertFalse(additional.get('custom'))



import json

from perimeterx import px_constants


def testing_mode_handling(ctx, config, request):
    result = {
        'px_cookies': ctx.px_cookies,
        'vid': ctx.vid,
        'ip': ctx.ip,
        'full_url': ctx.full_url,
        'px_cookie_hmac': ctx.cookie_hmac,
        'block_action': ctx.block_action,
        'http_method': ctx.http_method,
        'hostname': ctx.hostname,
        'headers': dict(ctx.headers),
        'user_agent': ctx.user_agent,
        'uri': ctx.uri,
        'is_made_s2s_api_call': True if ctx.s2s_call_reason != 'none' else False,
        'sensitive_route': ctx.sensitive_route,
        'decoded_px_cookie': ctx.decoded_cookie,
        'cookie_origin': ctx.cookie_origin,
        'http_version': ctx.http_version,
        's2s_call_reason': ctx.s2s_call_reason,
        'block_reason': ctx.block_reason,
        'module_mode': 1 if config.module_mode is px_constants.MODULE_MODE_BLOCKING else 0,
        'score': ctx.score,
        'risk_rtt': ctx.risk_rtt,
        'uuid': ctx.uuid,
        'pxde': ctx.pxde,
        'pxde_verified': ctx.pxde_verified,
    }

    if ctx.original_uuid:
        result['original_uuid'] = ctx.original_uuid
    if ctx.original_token_error:
        result['original_token_error'] = ctx.original_token_error

    return json.dumps(result), {'Content-Type': 'application/json'}, '200 OK'



import os
import unittest

from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import Request

from perimeterx.px_blocker import PXBlocker
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext


class Test_PXBlocker(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.config = PxConfig({'app_id': 'PXfake_app_id'})
        cls.headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'content-length': '100'}

    def test_is_json_response(self):
        px_blocker = PXBlocker()
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/init.js')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        self.assertFalse(px_blocker.is_json_response(context))
        context.headers['Accept'] = 'application/json'
        self.assertTrue(px_blocker.is_json_response(context))

    def test_handle_blocking(self):
        px_blocker = PXBlocker()
        vid = 'bf619be8-94be-458a-b6b1-ee81f154c282'
        px_uuid = '8712cef7-bcfa-4bb6-ae99-868025e1908a'

        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/init.js')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        context.vid = vid
        context.uuid = px_uuid
        px_config = PxConfig({'app_id': 'PXfake_app_id'})
        message, _, _ = px_blocker.handle_blocking(context, px_config)
        working_dir = os.path.dirname(os.path.realpath(__file__))
        with open(working_dir + '/px_blocking_messages/blocking.txt', 'r') as myfile:
            blocking_message = myfile.read()

        self.assertEqual(message, blocking_message)

    def test_handle_ratelimit(self):
        px_blocker = PXBlocker()
        vid = 'bf619be8-94be-458a-b6b1-ee81f154c282'
        px_uuid = '8712cef7-bcfa-4bb6-ae99-868025e1908a'
        config = PxConfig({'app_id': 'PXfake_app_id'})
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/init.js')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        context.vid = vid
        context.uuid = px_uuid
        context.block_action = 'r'
        message, _, _ = px_blocker.handle_blocking(context, config)
        blocking_message = None
        working_dir = os.path.dirname(os.path.realpath(__file__))
        with open(working_dir + '/px_blocking_messages/ratelimit.txt', 'r') as myfile:
            blocking_message = myfile.read()
        self.assertEqual(message, blocking_message)

    def test_handle_challenge(self):
        px_blocker = PXBlocker()
        vid = 'bf619be8-94be-458a-b6b1-ee81f154c282'
        px_uuid = '8712cef7-bcfa-4bb6-ae99-868025e1908a'
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/init.js')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        context.vid = vid
        context.uuid = px_uuid
        context.block_action = 'j'
        context.block_action_data = 'Bla'

        message, _, _ = px_blocker.handle_blocking(context, self.config)
        blocking_message = 'Bla'
        self.assertEqual(message, blocking_message)

    def test_prepare_properties(self):
        px_blocker = PXBlocker()
        vid = 'bf619be8-94be-458a-b6b1-ee81f154c282'
        px_uuid = '8712cef7-bcfa-4bb6-ae99-868025e1908a'
        builder = EnvironBuilder(headers=self.headers, path='/fake_app_id/init.js')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, self.config)
        context.vid = vid
        context.uuid = px_uuid
        message = px_blocker.prepare_properties(context, self.config)
        expected_message = {'blockScript': '/fake_app_id/captcha/captcha.js?a=&u=8712cef7-bcfa-4bb6-ae99-868025e1908a&v=bf619be8-94be-458a-b6b1-ee81f154c282&m=0',
                            'vid': 'bf619be8-94be-458a-b6b1-ee81f154c282',
                            'jsRef': '',
                            'hostUrl': '/fake_app_id/xhr',
                            'customLogo': '',
                            'appId': 'PXfake_app_id',
                            'uuid': '8712cef7-bcfa-4bb6-ae99-868025e1908a',
                            'jsClientSrc': '/fake_app_id/init.js',
                            'firstPartyEnabled': 'true',
                            'refId': '8712cef7-bcfa-4bb6-ae99-868025e1908a',
                            'cssRef': '',
                            'altBlockScript': '//captcha.px-cloud.net/PXfake_app_id/captcha.js?a=&u=8712cef7-bcfa-4bb6-ae99-868025e1908a&v=bf619be8-94be-458a-b6b1-ee81f154c282&m=0'}
        self.assertDictEqual(message, expected_message)
        expected_message['blockScript'] = '/fake_app/captcha/captcha.js?a=&u=8712cef7-bcfa-4bb6-ae99-868025e1908a&v=bf619be8-94be-458a-b6b1-ee81f154c282&m=0'
        self.assertNotEqual(message, expected_message)








import re
from time import gmtime, time

from perimeterx import px_constants


def merge_two_dicts(x, y):
    z = x.copy()  # start with x's keys and values
    z.update(y)  # modifies z with y's keys and values & returns None
    return z


def handle_proxy_headers(filtered_headers, ip, is_gae):
    if 'x-px-gae' in filtered_headers.keys():
        del filtered_headers['x-px-gae']
    for item in filtered_headers.copy().keys():
        if item.upper() == px_constants.FIRST_PARTY_FORWARDED_FOR:
            filtered_headers[item] = ip
        else:
            filtered_headers[px_constants.FIRST_PARTY_FORWARDED_FOR] = ip
    if is_gae:
        filtered_headers["x-px-gae"] = "true"
    return filtered_headers


def is_static_file(ctx):
    uri = ctx.uri
    static_extensions = ['.css', '.bmp', '.tif', '.ttf', '.docx', '.woff2', '.js', '.pict', '.tiff', '.eot',
                         '.xlsx', '.jpg', '.csv', '.eps', '.woff', '.xls', '.jpeg', '.doc', '.ejs', '.otf', '.pptx',
                         '.gif', '.pdf', '.swf', '.svg', '.ps', '.ico', '.pls', '.midi', '.svgz', '.class', '.png',
                         '.ppt', '.mid', 'webp', '.jar', '.json']

    for ext in static_extensions:
        if uri.endswith(ext):
            return True
    return False


custom_param_pattern = re.compile('(^(custom_param\d|custom_param10)$)')


def prepare_custom_params(config, dict_to_add):
    custom_params = {
        'custom_param1': '',
        'custom_param2': '',
        'custom_param3': '',
        'custom_param4': '',
        'custom_param5': '',
        'custom_param6': '',
        'custom_param7': '',
        'custom_param8': '',
        'custom_param9': '',
        'custom_param10': ''
    }
    if config.enrich_custom_parameters:
        risk_custom_params = config.enrich_custom_parameters(custom_params)
        if risk_custom_params:
            for param in risk_custom_params:
                if re.match(custom_param_pattern, param) and risk_custom_params[param] is not '':
                    dict_to_add[param] = risk_custom_params[param]


weekdayname = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']

monthname = [None,
             'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
             'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec']


def getExpiryDate(future=31536000):
    now = time()
    year, month, day, hh, mm, ss, wd, y, z = gmtime(now + future)
    return "%s, %02d %3s %4d %02d:%02d:%02d GMT" % \
           (weekdayname[wd], day, monthname[month], year, hh, mm, ss)



import logging

class Logger(object):
    def __init__(self, debug, app_id):
        self.debug_mode = debug
        self.app_id = app_id

        # Setup logger
        self.logger = logging.getLogger(__name__)
        handler = logging.StreamHandler()
        formatter = logging.Formatter('[PerimeterX %(levelname)s][{}]: %(message)s'.format(self.app_id))
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.logger.setLevel(logging.DEBUG)

    def debug(self, message):
        if self.debug_mode:
            self.logger.debug(message)

    def error(self, message):
        self.logger.error(message)



from perimeterx.px_cookie import PxCookie


class PxCookieV1(PxCookie):

    def __init__(self, config, raw_cookie):
        self._config = config
        self._logger = config.logger
        self._raw_cookie = raw_cookie

    def get_score(self):
        return self.decoded_cookie['s']['b']

    def get_hmac(self):
        return self.decoded_cookie['h']

    def get_action(self):
        return 'c'

    def is_cookie_format_valid(self):
        c = self.decoded_cookie
        return 't' in c and 'v' in c and 'u' in c and "s" in c and 'a' in c['s'] and 'h' in c

    def is_secured(self, user_agent, ip):
        c = self.decoded_cookie
        base_hmac = str(self.get_timestamp()) + str(c['s']['a']) + str(self.get_score()) + self.get_uuid() + self.get_vid()
        hmac_with_ip = base_hmac + ip + user_agent
        hmac_without_ip = base_hmac + user_agent

        return self.is_cookie_valid(hmac_without_ip) or self.is_cookie_valid(hmac_with_ip)





import time

from werkzeug.wrappers import Request

from perimeterx import px_activities_client
from perimeterx import px_utils
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext
from perimeterx.px_request_verifier import PxRequestVerifier


class PerimeterX(object):
    def __init__(self, app, config=None):
        self.app = app
        # merging user's defined configurations with the default one
        px_config = PxConfig(config)
        logger = px_config.logger

        if not px_config.app_id:
            logger.error('Unable to initialize module, missing mandatory configuration: app_id')
            raise ValueError('PX App ID is missing')

        if not px_config.auth_token:
            logger.error('Unable to initialize module, missing mandatory configuration: auth_token')
            raise ValueError('PX Auth Token is missing')

        if not px_config.cookie_key:
            logger.error('Unable to initialize module, missing mandatory configuration: px_cookie')
            raise ValueError('PX Cookie Key is missing')

        self.reverse_proxy_prefix = px_config.app_id[2:].lower()
        self._config = px_config
        self._request_verifier = PxRequestVerifier(px_config)
        px_activities_client.init_activities_configuration(px_config)

    def __call__(self, environ, start_response):
        context = None
        try:
            start = time.time()

            if environ.get('px_disable_request'):
                return self.app(environ, start_response)

            request = Request(environ)
            context, verified_response = self.verify(request)
            pxhd_callback = create_custom_pxhd_callback(context, start_response)
            self._config.logger.debug("PerimeterX Enforcer took: {} ms".format((time.time() - start) * 1000))
            px_activities_client.send_activities_in_thread()
            if verified_response is True:
                return self.app(environ, pxhd_callback)

            return verified_response(environ, pxhd_callback)

        except Exception as err:
            self._config.logger.error("Caught exception, passing request. Exception: {}".format(err))
            if context:
                self.report_pass_traffic(context)
            else:
                self.report_pass_traffic(PxContext(Request({}), self._config))
            return self.app(environ, start_response)

    def verify(self, request):
        config = self.config
        logger = config.logger
        logger.debug("Starting request verification {}".format(request.path))
        ctx = None
        try:
            if not config._module_enabled:
                logger.debug('Request will not be verified, module is disabled')
                return ctx, True
            ctx = PxContext(request, config)
            return ctx, self._request_verifier.verify_request(ctx, request)
        except Exception as err:
            logger.error("Caught exception in verify, passing request. Exception: {}".format(err))
            if ctx:
                self.report_pass_traffic(ctx)
            else:
                self.report_pass_traffic(PxContext(Request({}), config))
            return ctx, True

    def report_pass_traffic(self, ctx):
        px_activities_client.send_page_requested_activity(ctx, self.config)

    def report_block_traffic(self, ctx):
        px_activities_client.send_block_activity(ctx, self.config)

    @property
    def config(self):
        return self._config


def create_custom_pxhd_callback(context, start_response):
    def custom_start_response(status, headers, exc_info=None):
        if not context.pxhd or (context.response_pxhd and context.pxhd != context.response_pxhd):
            expiry = px_utils.getExpiryDate()
            headers.append(('Set-Cookie', "_pxhd={}; path=/; expires={}; ".format(context.response_pxhd, expiry)))
        return start_response(status, headers, exc_info)

    return custom_start_response



from perimeterx.px_config import PxConfig
import unittest
from perimeterx import px_constants
class TestPXConfig(unittest.TestCase):

    def test_constructor(self):
        config_dict = {'app_id': 'PXfake_app_id', 'debug_mode': True, 'module_mode': px_constants.MODULE_MODE_BLOCKING}
        config = PxConfig(config_dict)
        self.assertEqual(config._monitor_mode, 1)
        self.assertEqual(config.debug_mode, True)
        self.assertEqual(config.server_host, 'sapi-pxfake_app_id.perimeterx.net')
        self.assertEqual(config.collector_host, 'collector-pxfake_app_id.perimeterx.net')


import json

from perimeterx.px_cookie import PxCookie


class PxDataEnrichmentCookie(PxCookie):

    def __init__(self, config):
        super(PxDataEnrichmentCookie, self).__init__(config)
        self._is_valid = False
        self._payload = {}
        self._hmac = ''
        self._decoded_cookie = ''

    def from_raw_cookie(self, raw_cookie):
        if not raw_cookie:
            return

        spliced_cookie = raw_cookie.split(':')
        if len(spliced_cookie) != 2:
            return

        self._hmac = spliced_cookie[0]
        self._raw_cookie = spliced_cookie[1]
        self._is_valid = self.is_cookie_valid(self._raw_cookie)

        self._decoded_cookie = self.decode_cookie()
        try:
            self._payload = json.loads(self._decoded_cookie)
        except Exception as err:
            msg = 'PxDataEnrichmentCookie[__init__] failed to decode/unserialize data enrichment payload: %s' % err
            self._logger.debug(msg)

    @property
    def is_valid(self):
        return self._is_valid

    @is_valid.setter
    def is_valid(self, is_valid):
        self._is_valid = is_valid

    @property
    def payload(self):
        return self._payload

    @payload.setter
    def payload(self, payload):
        self._payload = payload



#!/usr/bin/env python

from setuptools import setup, find_packages

version = 'v1.2.0'
setup(name='perimeterx-python-3-wsgi',
      version=version,
      license='MIT',
      description='PerimeterX Python 3 WSGI middleware',
      author='Johnny Tordgeman',
      author_email='johnny@perimeterx.com',
      url='https://github.com/PerimeterX/perimeterx-python-3-wsgi',
      download_url='https://github.com/PerimeterX/perimeterx-python-3-wsgi/tarball/' + version,
      packages=find_packages(exclude=['dev', 'test*']),
      package_data={'perimeterx': ['templates/*', "*.json"]},
      install_requires=['pystache>=0.5.4', 'requests>=2.22.0', 'setuptools==41.4.0', 'Werkzeug==0.16.0', 'pycryptodome>=3.9.0'],
      classifiers=['Intended Audience :: Developers',
                   'Programming Language :: Python :: 3.7'])



import re

from requests.structures import CaseInsensitiveDict

from perimeterx.px_constants import *
from perimeterx.px_data_enrichment_cookie import PxDataEnrichmentCookie

class PxContext(object):

    def __init__(self, request, config):
        logger = config.logger
        # Default values
        px_cookies = {}
        request_cookie_names = []
        cookie_origin = "cookie"
        vid = ''
        enforcer_vid_source = ''
        pxhd = ''
        data_enrichment = PxDataEnrichmentCookie(config)
        if not hasattr(request, 'headers'):
            request_headers = {}
        else:
            request_headers = request.headers
        headers = generate_context_headers(request_headers, config.sensitive_headers)
        original_token = ''
        mobile_header = headers.get(MOBILE_SDK_HEADER)
        if mobile_header is None:
            cookies = request.cookies
            for cookie_key in cookies:
                cookie_value = cookies[cookie_key]
                request_cookie_names.append(cookie_key)
                if cookie_key == PREFIX_PX_COOKIE_V1 or cookie_key == PREFIX_PX_COOKIE_V3:
                    logger.debug('Found cookie prefix:' + cookie_key)
                    px_cookies[cookie_key] = cookie_value
                elif cookie_key == PREFIX_PX_DATA_ENRICHMENT:
                    data_enrichment.from_raw_cookie(cookie_value)
                elif cookie_key == PREFIX_PXVID or cookie_key == "_" + PREFIX_PXVID :
                    vid = cookie_value
                    enforcer_vid_source = 'vid_cookie'
                elif cookie_key == PREFIX_PXHD:
                    pxhd = cookie_value
        else:
            cookie_origin = "header"
            original_token = headers.get(MOBILE_SDK_ORIGINAL_HEADER)
            logger.debug('Mobile SDK token detected')
            cookie_name, cookie = self.get_token_object(config, mobile_header)
            px_cookies[cookie_name] = cookie

        user_agent = request.user_agent.string
        uri = request.path
        full_url = request.url
        hostname = request.host
        sensitive_route = sum(1 for _ in filter(lambda sensitive_route_item:  re.search(sensitive_route_item, uri), config.sensitive_routes_regex)) > 0 or sum(1 for _ in filter(lambda sensitive_route_item: uri.startswith(sensitive_route_item), config.sensitive_routes)) > 0
        whitelist_route = sum(1 for _ in filter(lambda whitelist_route_item: re.search(whitelist_route_item, uri), config.whitelist_routes_regex)) > 0 or sum(1 for _ in filter(lambda whitelist_route_item: uri.startswith(whitelist_route_item), config.whitelist_routes)) > 0
        enforced_route = sum(1 for _ in filter(lambda enforced_route_item: re.search(enforced_route_item, uri), config.enforced_specific_routes_regex)) > 0 or sum(1 for _ in filter(lambda enforced_route_item: uri.startswith(enforced_route_item), config.enforced_specific_routes)) > 0
        monitored_route = not enforced_route and (sum(1 for _ in filter(lambda monitored_route_item: re.search(monitored_route_item, uri), config.monitored_specific_routes_regex)) > 0 or sum(1 for _ in filter(lambda monitored_route_item: uri.startswith(monitored_route_item), config.monitored_specific_routes)) > 0)

        protocol_split = request.environ.get('SERVER_PROTOCOL', '').split('/')
        if protocol_split[0].startswith('HTTP'):
            self._http_protocol = protocol_split[0].lower() + '://'
        if len(protocol_split) > 1:
            self._http_version = protocol_split[1]

        self._headers = headers
        self._http_method = request.environ.get('REQUEST_METHOD')
        self._user_agent = user_agent
        self._full_url = full_url
        self._uri = uri
        self._hostname = hostname
        self._px_cookies = px_cookies
        self._cookie_names = request_cookie_names
        self._risk_rtt = 0
        self._ip = self.extract_ip(config, request)
        self._vid = vid
        self._pxhd = pxhd
        self._response_pxhd = ''
        self._enforcer_vid_source = enforcer_vid_source
        self._uuid = ''
        self._query_params = request.query_string.decode("utf-8")
        self._sensitive_route = sensitive_route
        self._whitelist_route = whitelist_route
        self._enforced_route = enforced_route
        self._monitored_route = monitored_route
        self._s2s_call_reason = 'none'
        self._cookie_origin = cookie_origin
        self._is_mobile = cookie_origin == "header"
        self._score = -1
        self._block_reason = ''
        self._decoded_cookie = ''
        self._block_action = ''
        self._block_action_data = ''
        self._pass_reason = 'cookie'
        self._cookie_hmac = ''
        self._px_cookie_raw = ''
        self._original_token_error = ''
        self._original_uuid = ''
        self._decoded_original_token = ''
        self._original_token = original_token
        self._pxde = data_enrichment.payload if data_enrichment else ''
        self._pxde_verified = data_enrichment.is_valid if data_enrichment else False
        logger.debug('Request context created successfully')

    def get_token_object(self, config, token):
        logger = config.logger
        sliced_token = token.split(":", 1)
        if len(sliced_token) > 1:
            key = sliced_token.pop(0)
            if key == PREFIX_PX_TOKEN_V1 or key == PREFIX_PX_TOKEN_V3:
                logger.debug('Found token prefix: {}'.format(key))
                return key, sliced_token[0]
        return PREFIX_PX_TOKEN_V3, token

    def extract_ip(self, config, request):
        request_headers = request.headers
        ip = request_headers.get('X-FORWARDED-FOR') if request_headers.get('X-FORWARDED-FOR') else request.remote_addr
        ip_headers = config.ip_headers
        logger = config.logger
        if ip_headers:
            try:
                for ip_header in ip_headers:
                    if request_headers.get(ip_header):
                        return request_headers.get(ip_header)
            except:
                logger.debug('Failed to use IP_HEADERS from config')
        if config.get_user_ip:
            ip = config.get_user_ip
        return ip

    @property
    def enforced_route(self):
        return self._enforced_route

    @enforced_route.setter
    def enforced_route(self, enforced_route):
        self._enforced_route = enforced_route

    @property
    def headers(self):
        return self._headers

    @headers.setter
    def headers(self, headers):
        self._headers = headers

    @property
    def http_method(self):
        return self._http_method

    @http_method.setter
    def http_method(self, http_method):
        self._http_method = http_method

    @property
    def http_version(self):
        return self._http_version

    @http_version.setter
    def http_version(self, http_version):
        self._http_version = http_version

    @property
    def http_protocol(self):
        return self._http_protocol

    @http_protocol.setter
    def http_protocol(self, http_protocol):
        self._http_protocol = http_protocol

    @property
    def user_agent(self):
        return self._user_agent

    @user_agent.setter
    def user_agent(self, user_agent):
        self._user_agent = user_agent

    @property
    def full_url(self):
        return self._full_url

    @full_url.setter
    def full_url(self, full_url):
        self._full_url = full_url

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, uri):
        self._uri = uri

    @property
    def hostname(self):
        return self._hostname

    @hostname.setter
    def hostname(self, hostname):
        self._hostname = hostname

    @property
    def px_cookies(self):
        return self._px_cookies

    @px_cookies.setter
    def px_cookies(self, px_cookies):
        self._px_cookies = px_cookies

    @property
    def cookie_names(self):
        return self._cookie_names

    @cookie_names.setter
    def cookie_names(self, cookie_names):
        self._cookie_names = cookie_names

    @property
    def risk_rtt(self):
        return self._risk_rtt

    @risk_rtt.setter
    def risk_rtt(self, risk_rtt):
        self._risk_rtt = risk_rtt

    @property
    def ip(self):
        return self._ip

    @ip.setter
    def ip(self, ip):
        self._ip = ip

    @property
    def vid(self):
        return self._vid

    @vid.setter
    def vid(self, vid):
        self._vid = vid

    @property
    def query_params(self):
        return self._query_params

    @query_params.setter
    def query_params(self, query_params):
        self._query_params = query_params

    @property
    def sensitive_route(self):
        return self._sensitive_route

    @sensitive_route.setter
    def sensitive_route(self, sensitive_route):
        self._sensitive_route = sensitive_route

    @property
    def whitelist_route(self):
        return self._whitelist_route

    @whitelist_route.setter
    def whitelist_route(self, whitelist_route):
        self._whitelist_route = whitelist_route

    @property
    def monitored_route(self):
        return self._monitored_route

    @monitored_route.setter
    def monitored_route(self, monitored_route):
        self._monitored_route = monitored_route

    @property
    def s2s_call_reason(self):
        return self._s2s_call_reason

    @s2s_call_reason.setter
    def s2s_call_reason(self, s2s_call_reason):
        self._s2s_call_reason = s2s_call_reason

    @property
    def cookie_origin(self):
        return self._cookie_origin

    @cookie_origin.setter
    def cookie_origin(self, cookie_origin):
        self._cookie_origin = cookie_origin

    @property
    def original_token(self):
        return self._original_token

    @original_token.setter
    def original_token(self, original_token):
        self._original_token = original_token

    @property
    def is_mobile(self):
        return self._is_mobile

    @is_mobile.setter
    def is_mobile(self, is_mobile):
        self._is_mobile = is_mobile

    @property
    def score(self):
        return self._score

    @score.setter
    def score(self, score):
        self._score = score

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, uuid):
        self._uuid = uuid

    @property
    def block_reason(self):
        return self._block_reason

    @block_reason.setter
    def block_reason(self, block_reason):
        self._block_reason = block_reason

    @property
    def decoded_cookie(self):
        return self._decoded_cookie

    @decoded_cookie.setter
    def decoded_cookie(self, decoded_cookie):
        self._decoded_cookie = decoded_cookie

    @property
    def block_action(self):
        return self._block_action

    @block_action.setter
    def block_action(self, block_action):
        self._block_action = block_action

    @property
    def block_action_data(self):
        return self._block_action_data

    @block_action_data.setter
    def block_action_data(self, block_action_data):
        self._block_action_data = block_action_data

    @property
    def pass_reason(self):
        return self._pass_reason

    @pass_reason.setter
    def pass_reason(self, pass_reason):
        self._pass_reason = pass_reason

    @property
    def cookie_hmac(self):
        return self._cookie_hmac

    @cookie_hmac.setter
    def cookie_hmac(self, cookie_hmac):
        self._cookie_hmac = cookie_hmac

    @property
    def px_cookie_raw(self):
        return self._px_cookie_raw

    @px_cookie_raw.setter
    def px_cookie_raw(self, px_cookie_raw):
        self._px_cookie_raw = px_cookie_raw

    @property
    def original_token_error(self):
        return self._original_token_error

    @original_token_error.setter
    def original_token_error(self, original_token_error):
        self._original_token_error = original_token_error

    @property
    def original_uuid(self):
        return self._original_uuid

    @original_uuid.setter
    def original_uuid(self, original_uuid):
        self._original_uuid = original_uuid

    @property
    def decoded_original_token(self):
        return self._decoded_original_token

    @decoded_original_token.setter
    def decoded_original_token(self, decoded_original_token):
        self._decoded_original_token = decoded_original_token

    @property
    def pxde(self):
        return self._pxde

    @pxde.setter
    def pxde(self, pxde):
        self._pxde = pxde

    @property
    def pxde_verified(self):
        return self._pxde_verified

    @pxde_verified.setter
    def pxde_verified(self, pxde_verified):
        self._pxde_verified = pxde_verified

    @property
    def enforcer_vid_source(self):
        return self._enforcer_vid_source

    @enforcer_vid_source.setter
    def enforcer_vid_source(self, enforcer_vid_source):
        self._enforcer_vid_source = enforcer_vid_source

    @property
    def pxhd(self):
        return self._pxhd

    @pxhd.setter
    def pxhd(self, pxhd):
        self._pxhd = pxhd

    @property
    def response_pxhd(self):
        return self._response_pxhd

    @response_pxhd.setter
    def response_pxhd(self, response_pxhd):
        self._response_pxhd = response_pxhd


def generate_context_headers(request_headers, sensitive_headers):
    headers = CaseInsensitiveDict()
    for header_name, header_value in request_headers:
        if header_name.lower() not in sensitive_headers:
            headers[header_name] = header_value
    return headers



import unittest
from perimeterx.px_data_enrichment_cookie import PxDataEnrichmentCookie
from perimeterx.px_config import PxConfig
import base64
import json
import hmac
import hashlib


class TestPXDataEnrichmentCookie(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        config = PxConfig({'app_id': 'fake_app_id', 'cookie_key': 'test_key'})
        cls.config = config

    def test_from_raw_cookie(self):
        data_enrichment_cookie = PxDataEnrichmentCookie(self.config)
        data = {'timestamp': 1398239283293}
        data_json_string = json.dumps(data)
        encoded_data = base64.b64encode(data_json_string.encode("utf-8"))
        hmac_hex = hmac.new(self.config.cookie_key.encode("utf-8"), encoded_data, hashlib.sha256).hexdigest()
        raw_cookie = hmac_hex  + ':' + encoded_data.decode("utf-8")
        data_enrichment_cookie.from_raw_cookie(raw_cookie)
        self.assertEqual(data_enrichment_cookie.payload, data)
        self.assertTrue(data_enrichment_cookie.is_valid)



import json
import unittest
import uuid

import mock
from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import Request

from perimeterx import px_api
from perimeterx.px_config import PxConfig
from perimeterx.px_context import PxContext


class Test_PXApi(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.headers = {'X-FORWARDED-FOR': '127.0.0.1',
                       'remote-addr': '127.0.0.1',
                       'content-length': '100'}

    def enrich_custom_parameters(self, params):
        params['custom_param1'] = '1'
        params['custom_param2'] = '5'
        params['custom'] = '6'
        return params

    def test_prepare_risk_body(self):
        config = PxConfig({'app_id': 'app_id', 'enrich_custom_parameters': self.enrich_custom_parameters})
        builder = EnvironBuilder(headers=self.headers, path='/')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, config)
        context.s2s_call_reason = 'no_cookie'
        body = px_api.prepare_risk_body(context, config)
        self.assertEqual(body['additional'].get('custom_param1'), '1')
        self.assertEqual(body['additional'].get('custom_param2'), '5')
        self.assertFalse(body['additional'].get('custom'))

    def test_send_risk_request(self):
        config = PxConfig({'app_id': 'app_id',
                           'enrich_custom_parameters': self.enrich_custom_parameters,
                           'auth_token': 'auth'})
        builder = EnvironBuilder(headers=self.headers, path='/test_path')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, config)
        uuid_val = str(uuid.uuid4())
        response = ResponseMock({'score': 100, 'uuid': uuid_val, 'action': 'c'})
        with mock.patch('perimeterx.px_httpc.send', return_value=response):
            response = px_api.send_risk_request(context, config)
            self.assertEqual({'action': 'c', 'score': 100, 'uuid': uuid_val}, response)

    def test_verify(self):
        config = PxConfig({'app_id': 'app_id',
                           'enrich_custom_parameters': self.enrich_custom_parameters,
                           'auth_token': 'auth'})
        builder = EnvironBuilder(headers=self.headers, path='/test_path')

        env = builder.get_environ()
        request = Request(env)
        context = PxContext(request, config)
        uuid_val = str(uuid.uuid4())
        data_enrichment = {'timestamp': 10033200222}
        response = {'score': 100, 'uuid': uuid_val, 'action': 'c', 'data_enrichment': data_enrichment}
        with mock.patch('perimeterx.px_api.send_risk_request', return_value=response):
            api_response = px_api.verify(context, config)
            self.assertEqual('s2s_high_score', context.block_reason)
            self.assertEqual('c', context.block_action)
            self.assertTrue(api_response)
            self.assertEqual(data_enrichment, context.pxde)



class ResponseMock(object):
    def __init__(self, dict):
        self.content = json.dumps(dict)



import base64

from werkzeug.wrappers import Response

from perimeterx import px_constants
from perimeterx import px_httpc
from perimeterx import px_utils
import os

hoppish = {'connection', 'keep-alive', 'proxy-authenticate',
           'proxy-authorization', 'te', 'trailers', 'transfer-encoding',
           'upgrade'
           }


def delete_extra_headers(filtered_headers):
    if 'content-length' in filtered_headers.keys():
        del filtered_headers['content-length']
    if 'content-type' in filtered_headers.keys():
        del filtered_headers['content-type']


class PxProxy(object):
    def __init__(self, config):
        self._logger = config.logger
        self.is_gae = os.environ.get('SERVER_SOFTWARE','').startswith('Google')
        reverse_app_id = config.app_id[2:]
        self.client_reverse_prefix = '/{}/{}'.format(reverse_app_id, px_constants.CLIENT_FP_PATH).lower()
        self.xhr_reverse_prefix = '/{}/{}'.format(reverse_app_id, px_constants.XHR_FP_PATH).lower()
        self.captcha_reverse_prefix = '/{}/{}'.format(reverse_app_id, px_constants.CAPTCHA_FP_PATH).lower()

    def should_reverse_request(self, uri):
        uri = uri.lower()
        if uri.startswith(self.client_reverse_prefix) or uri.startswith(self.xhr_reverse_prefix) or uri.startswith(
                self.captcha_reverse_prefix):
            return True
        return False

    def handle_reverse_request(self, config, ctx, body):
        uri = ctx.uri.lower()

        if uri.startswith(self.client_reverse_prefix):
            status, headers, data = self.send_reverse_client_request(config=config, ctx=ctx)
        elif uri.startswith(self.xhr_reverse_prefix):
            status, headers, data = self.send_reverse_xhr_request(config=config, ctx=ctx, body=body)
        elif uri.startswith(self.captcha_reverse_prefix):
            status, headers, data = self.send_reverse_captcha_request(config=config, ctx=ctx)
        px_response = Response(data)
        px_response.headers = filter_hop_by_hop_headers(headers)
        px_response.status = status
        return px_response

    def send_reverse_client_request(self, config, ctx):
        if not config.first_party:
            headers = {'Content-Type': 'application/javascript'}
            return '200 OK', headers, ''

        client_request_uri = '/{}/main.min.js'.format(config.app_id)
        msg = 'Forwarding request from {} to client at {}{}'
        self._logger.debug(msg.format(ctx.uri.lower(), px_constants.CLIENT_HOST, client_request_uri))

        headers = {'host': px_constants.CLIENT_HOST,
                   px_constants.FIRST_PARTY_HEADER: '1',
                   px_constants.ENFORCER_TRUE_IP_HEADER: ctx.ip}
        filtered_headers = px_utils.handle_proxy_headers(ctx.headers, ctx.ip, self.is_gae)
        filtered_headers = px_utils.merge_two_dicts(filtered_headers, headers)
        delete_extra_headers(filtered_headers)
        px_response = px_httpc.send(full_url=px_constants.CLIENT_HOST + client_request_uri, body='',
                                    headers=filtered_headers, config=config, method='GET')
        if self.is_gae:
            data = px_response.raw.read(decode_content=True)
        else:
            data = px_response.raw.read()
        headers = px_response.headers
        status = str(px_response.status_code) + ' ' + str(px_response.reason)
        return status, headers, data

    def send_reverse_xhr_request(self, config, ctx, body):
        uri = ctx.uri
        if not config.first_party or not config.first_party_xhr_enabled:
            body, response_headers = self.return_default_response(uri)
            return '200 OK', response_headers, body

        xhr_path_index = uri.find('/' + px_constants.XHR_FP_PATH)
        suffix_uri = uri[xhr_path_index + 4:]
        host = config.collector_host
        headers = {'host': host,
                   px_constants.FIRST_PARTY_HEADER: '1',
                   px_constants.ENFORCER_TRUE_IP_HEADER: ctx.ip}

        if ctx.vid is not None:
            headers['cookie'] = 'pxvid=' + ctx.vid

        filtered_headers = px_utils.handle_proxy_headers(ctx.headers, ctx.ip, self.is_gae)
        filtered_headers = px_utils.merge_two_dicts(filtered_headers, headers)
        msg = 'Forwarding request from {} to client at {}{}'
        self._logger.debug(msg.format(ctx.uri.lower(), host, suffix_uri))
        px_response = px_httpc.send(full_url=host + suffix_uri + "?" + ctx.query_params, body=body,
                                    headers=filtered_headers, config=config, method=ctx.http_method)

        if px_response.status_code >= 400:
            data, response_headers = self.return_default_response(uri)
            self._logger.debug('Error reversing the http call: {}'.format(px_response.reason))
            return '200 OK', response_headers, data
        data = px_response.content
        headers = px_response.headers
        status = str(px_response.status_code) + ' ' + str(px_response.reason)
        return status, headers, data

    def send_reverse_captcha_request(self, config, ctx):
        if not config.first_party:
            status = '200 OK'
            response_headers = {'Content-Type': 'application/javascript'}
            return status, response_headers, ''
        uri = '/{}{}?{}'.format(config.app_id, ctx.uri.lower().replace(self.captcha_reverse_prefix, ''),
                                ctx.query_params)
        host = px_constants.CAPTCHA_HOST

        headers = {'host': px_constants.CAPTCHA_HOST,
                   px_constants.FIRST_PARTY_HEADER: '1',
                   px_constants.ENFORCER_TRUE_IP_HEADER: ctx.ip}
        filtered_headers = px_utils.handle_proxy_headers(ctx.headers, ctx.ip, self.is_gae)
        filtered_headers = px_utils.merge_two_dicts(filtered_headers, headers)
        delete_extra_headers(filtered_headers)
        self._logger.debug('Forwarding request from {} to client at {}{}'.format(ctx.uri.lower(), host, uri))
        px_response = px_httpc.send(full_url=host + uri, body='',
                                    headers=filtered_headers, config=config, method='GET')
        if self.is_gae:
            data = px_response.raw.read(decode_content=True)
        else:
            data = px_response.raw.read()
        headers = px_response.headers
        status = str(px_response.status_code) + ' ' + str(px_response.reason)
        return status, headers, data

    def return_default_response(self, uri):
        if 'gif' in uri.lower():
            headers = {'Content-Type': 'image/gif'}
            body = base64.b64decode(px_constants.EMPTY_GIF_B64)
        else:
            headers = {'Content-Type': 'application/json'}
            body = {}
        return body, headers

def filter_hop_by_hop_headers(response_headers):
    headers = {}
    for key in response_headers.keys():
        if key.lower() not in hoppish:
            headers[key] = response_headers[key]
    return headers